from typing import Union
from fastapi import APIRouter, Header

from ._authenticate_gui_request import _authenticate_gui_request
from ...services.gui.create_job import create_job
from ..common import api_route_wrapper

from ....common.dendro_types import CreateJobRequest, CreateJobResponse


router = APIRouter()

# create job

@router.post("/jobs")
@api_route_wrapper
async def create_job_handler(
    data: CreateJobRequest,
    github_access_token: Union[str, None] = Header(None),
    dendro_api_key: Union[str, None] = Header(None)
) -> CreateJobResponse:
    # authenticate the request
    user_id = await _authenticate_gui_request(
        github_access_token=github_access_token,
        dendro_api_key=dendro_api_key,
        raise_on_not_authenticated=True
    )
    assert user_id

    # parse the request
    project_id = data.projectId
    processor_name = data.processorName
    input_files_from_request = data.inputFiles
    output_files_from_request = data.outputFiles
    input_parameters = data.inputParameters
    processor_spec = data.processorSpec
    batch_id = data.batchId
    dandi_api_key = data.dandiApiKey
    required_resources = data.requiredResources
    run_method = data.runMethod

    job_id = await create_job(
        project_id=project_id,
        processor_name=processor_name,
        input_files_from_request=input_files_from_request,
        output_files_from_request=output_files_from_request,
        input_parameters=input_parameters,
        processor_spec=processor_spec,
        batch_id=batch_id,
        user_id=user_id,
        dandi_api_key=dandi_api_key,
        required_resources=required_resources,
        run_method=run_method,
        pending_approval=True
    )

    return CreateJobResponse(
        jobId=job_id,
        success=True
    )



import numpy as np
import h5py


def NwbSorting(file):
    h5_file = h5py.File(file, 'r')

    # Load unit IDs
    ids: np.ndarray = h5_file['units']['id'][:] # type: ignore

    # Load spike times index
    spike_times_index: np.ndarray = h5_file['units']['spike_times_index'][:] # type: ignore

    # Load spike times
    spike_times: np.ndarray = h5_file['units']['spike_times'][:] # type: ignore

    units_dict = {}
    sampling_frequency = 30000 # TODO: get this from the NWB file
    for i in range(len(ids)):
        if i == 0:
            s = spike_times[0:spike_times_index[0]]
        else:
            s = spike_times[spike_times_index[i - 1]:spike_times_index[i]]
        units_dict[ids[i]] = (s * sampling_frequency).astype(np.int32)
    sorting = _numpy_sorting_from_dict([units_dict], sampling_frequency=sampling_frequency)
    return sorting

def _numpy_sorting_from_dict(units_dict_list, *, sampling_frequency):
    import spikeinterface as si # type: ignore
    try:
        # different versions of spikeinterface
        # see: https://github.com/SpikeInterface/spikeinterface/issues/2083
        sorting = si.NumpySorting.from_dict( # type: ignore
            units_dict_list, sampling_frequency=sampling_frequency # type: ignore
        )
    except: # noqa
        sorting = si.NumpySorting.from_unit_dict( # type: ignore
            units_dict_list, sampling_frequency=sampling_frequency # type: ignore
        )
    return sorting



from dendro.api_helpers.services._crypto_keys import generate_keypair


def test_crypto_keys():
    public_key_hex, private_key_hex = generate_keypair() # this does checks internally
    assert public_key_hex
    assert private_key_hex



import os
import requests
from ._crypto_keys import _sign_message_str

dendro_url = os.getenv('DENDRO_URL', 'https://dendro.vercel.app')

_globals = {
    'test_client': None
}
def _use_api_test_client(test_client):
    _globals['test_client'] = test_client

def _compute_resource_get_api_request(*,
    url_path: str,
    compute_resource_id: str,
    compute_resource_private_key: str,
    _wrong_payload_for_testing: bool = False,
    _wrong_signature_for_testing: bool = False
):
    payload = url_path
    signature = _sign_message_str(payload, compute_resource_id, compute_resource_private_key)

    if _wrong_payload_for_testing:
        payload = 'wrong payload'
    if _wrong_signature_for_testing:
        signature = 'wrong signature'

    headers = {
        'compute-resource-id': compute_resource_id,
        'compute-resource-payload': payload,
        'compute-resource-signature': signature
    }

    test_client = _globals['test_client']
    if test_client is None:
        url = f'{dendro_url}{url_path}'
        client = requests
    else:
        assert url_path.startswith('/api')
        url = url_path
        client = test_client
    try:
        resp = client.get(url, headers=headers, timeout=60)
        resp.raise_for_status()
    except Exception as e:
        print(f'Error in compute resource get api request for {url}; {e}')
        raise
    return resp.json()

# not used right now
# def _compute_resource_post_api_request(*,
#     url_path: str,
#     compute_resource_id: str,
#     compute_resource_private_key: str,
#     data: dict
# ):
#     payload = url_path
#     signature = _sign_message_str(payload, compute_resource_id, compute_resource_private_key)

#     headers = {
#         'compute-resource-id': compute_resource_id,
#         'compute-resource-payload': payload,
#         'compute-resource-signature': signature
#     }

#     test_client = _globals['test_client']
#     if test_client is None:
#         url = f'{dendro_url}{url_path}'
#         client = requests
#     else:
#         assert url_path.startswith('/api')
#         url = url_path
#         client = test_client
#     try:
#         resp = client.post(url, headers=headers, json=data, timeout=60)
#         resp.raise_for_status()
#     except Exception as e:
#         print(f'Error in compute resource post api request for {url}; {e}')
#         raise
#     return resp.json()

def _compute_resource_put_api_request(*,
    url_path: str,
    compute_resource_id: str,
    compute_resource_private_key: str,
    data: dict
):
    payload = url_path
    signature = _sign_message_str(payload, compute_resource_id, compute_resource_private_key)

    headers = {
        'compute-resource-id': compute_resource_id,
        'compute-resource-payload': payload,
        'compute-resource-signature': signature
    }

    test_client = _globals['test_client']
    if test_client is None:
        url = f'{dendro_url}{url_path}'
        client = requests
    else:
        assert url_path.startswith('/api')
        url = url_path
        client = test_client
    try:
        resp = client.put(url, headers=headers, json=data, timeout=60)
        resp.raise_for_status()
    except Exception as e:
        print(f'Error in compute resource put api request for {url}; {e}')
        raise
    return resp.json()

def _processor_get_api_request(*,
    url_path: str,
    headers: dict
):
    test_client = _globals['test_client']
    if test_client is None:
        url = f'{dendro_url}{url_path}'
        client = requests
    else:
        assert url_path.startswith('/api')
        url = url_path
        client = test_client
    try:
        resp = client.get(url, headers=headers, timeout=60)
        resp.raise_for_status()
    except Exception as e:
        print(f'Error in processor get api request for {url}; {e}')
        raise
    return resp.json()

def _processor_put_api_request(*,
    url_path: str,
    headers: dict,
    data: dict
):
    test_client = _globals['test_client']
    if test_client is None:
        url = f'{dendro_url}{url_path}'
        client = requests
    else:
        assert url_path.startswith('/api')
        url = url_path
        client = test_client
    try:
        resp = client.put(url, headers=headers, json=data, timeout=60)
        resp.raise_for_status()
    except Exception as e:
        print(f'Error in processor put api request for {url}; {e}')
        raise
    return resp.json()

def _client_get_api_request(*,
    url_path: str
):
    test_client = _globals['test_client']
    if test_client is None:
        url = f'{dendro_url}{url_path}'
        client = requests
    else:
        assert url_path.startswith('/api')
        url = url_path
        client = test_client
    try:
        resp = client.get(url, timeout=60)
        resp.raise_for_status()
    except Exception as e:
        print(f'Error in client get api request for {url}; {e}')
        raise
    return resp.json()

def _client_post_api_request(*,
    url_path: str,
    data: dict,
    dendro_api_key: str
):
    headers = {
        'dendro-api-key': dendro_api_key
    }
    test_client = _globals['test_client']
    if test_client is None:
        url = f'{dendro_url}{url_path}'
        client = requests
    else:
        assert url_path.startswith('/api')
        url = url_path
        client = test_client
    try:
        resp = client.post(url, headers=headers, json=data, timeout=60)
        resp.raise_for_status()
    except Exception as e:
        print(f'Error in client post api request for {url}; {e}')
        raise
    return resp.json()

####################################################################################################
# The GUI API requests below are only here for use with pytest since the real GUI requests come from the browser using typescript

def _gui_get_api_request(*,
    url_path: str,
    github_access_token: str
):
    headers = {
        'github-access-token': github_access_token
    }
    test_client = _globals['test_client']
    if test_client is None:
        url = f'{dendro_url}{url_path}'
        client = requests
    else:
        assert url_path.startswith('/api')
        url = url_path
        client = test_client
    try:
        resp = client.get(url, headers=headers, timeout=60)
        resp.raise_for_status()
    except: # noqa E722
        print(f'Error in gui get api request for {url}')
        raise
    return resp.json()

def _gui_post_api_request(*,
    url_path: str,
    github_access_token: str,
    data: dict
):
    headers = {
        'github-access-token': github_access_token
    }
    test_client = _globals['test_client']
    if test_client is None:
        url = f'{dendro_url}{url_path}'
        client = requests
    else:
        assert url_path.startswith('/api')
        url = url_path
        client = test_client
    try:
        resp = client.post(url, headers=headers, json=data, timeout=60)
        resp.raise_for_status()
    except: # noqa E722
        print(f'Error in gui post api request for {url}')
        raise
    return resp.json()

def _gui_put_api_request(*,
    url_path: str,
    github_access_token: str,
    data: dict
):
    headers = {
        'github-access-token': github_access_token
    }
    test_client = _globals['test_client']
    if test_client is None:
        url = f'{dendro_url}{url_path}'
        client = requests
    else:
        assert url_path.startswith('/api')
        url = url_path
        client = test_client
    try:
        resp = client.put(url, headers=headers, json=data, timeout=60)
        resp.raise_for_status()
    except: # noqa E722
        print(f'Error in gui put api request for {url}')
        raise
    return resp.json()

def _gui_delete_api_request(*,
    url_path: str,
    github_access_token: str
):
    headers = {
        'github-access-token': github_access_token
    }
    test_client = _globals['test_client']
    if test_client is None:
        url = f'{dendro_url}{url_path}'
        client = requests
    else:
        assert url_path.startswith('/api')
        url = url_path
        client = test_client
    try:
        resp = client.delete(url, headers=headers, timeout=60)
        resp.raise_for_status()
    except: # noqa E722
        print(f'Error in gui delete api request for {url}')
        raise
    return resp.json()



import shutil
import tempfile
from dendro.compute_resource.register_compute_resource import register_compute_resource


def test_register_compute_resource():
    with TemporaryDirectory() as tmpdir:
        register_compute_resource(
            dir=tmpdir
        )

class TemporaryDirectory:
    """A context manager for temporary directories"""
    def __init__(self):
        self._dir = None
    def __enter__(self):
        self._dir = tempfile.mkdtemp()
        return self._dir
    def __exit__(self, exc_type, exc_value, traceback):
        if self._dir:
            shutil.rmtree(self._dir)



from .... import BaseModel
from fastapi import APIRouter
import aiohttp
from ...core.settings import get_settings


router = APIRouter()

class GithubAuthError(Exception):
    pass

# github auth
class GithubAuthResponse(BaseModel):
    access_token: str

@router.get("/github_auth/{code}")
async def github_auth(code) -> GithubAuthResponse:
    settings = get_settings()
    GITHUB_CLIENT_ID = settings.GITHUB_CLIENT_ID
    GITHUB_CLIENT_SECRET = settings.GITHUB_CLIENT_SECRET
    assert GITHUB_CLIENT_ID, 'Env var not set: VITE_GITHUB_CLIENT_ID'
    assert GITHUB_CLIENT_SECRET, 'Env var not set: GITHUB_CLIENT_SECRET'
    url = f'https://github.com/login/oauth/access_token?client_id={GITHUB_CLIENT_ID}&client_secret={GITHUB_CLIENT_SECRET}&code={code}'
    headers = {
        'accept': 'application/json'
    }
    async with aiohttp.ClientSession() as session:
        async with session.get(url, headers=headers) as resp:
            r = await resp.json()
            if 'access_token' in r:
                return GithubAuthResponse(access_token=r['access_token']) # pragma: no cover
            elif 'error' in r:
                raise GithubAuthError(f'Error in github oauth response: {r["error"]}')
            else:
                raise Exception('No access_token in github oauth response.') # pragma: no cover



import pytest
import os


@pytest.mark.asyncio
@pytest.mark.api
async def test_github_auth():
    from dendro.api_helpers.routers.gui.github_auth_routes import github_auth, GithubAuthError
    old_env = os.environ.copy()
    try:
        os.environ['VITE_GITHUB_CLIENT_ID'] = 'test-client-id'
        os.environ['GITHUB_CLIENT_SECRET'] = 'test-client-secret'
        with pytest.raises(GithubAuthError):
            await github_auth(code='test-code')
    finally:
        os.environ = old_env



from typing import Optional, Dict, Any
import boto3
from boto3.session import Config
import json


async def _get_signed_upload_url(*,
    bucket_uri: str,
    bucket_credentials: str,
    object_key: str,
    size: Optional[int] = None
):
    creds = json.loads(bucket_credentials)
    access_key_id = creds['accessKeyId']
    secret_access_key = creds['secretAccessKey']
    endpoint = creds.get('endpoint', None)

    region_name = _get_region_name_from_uri(bucket_uri)
    bucket_name = _get_bucket_name_from_uri(bucket_uri)

    s3_client = boto3.client(
        's3',
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key,
        endpoint_url=endpoint,
        region_name=region_name,
        config=Config(signature_version='s3v4')
    )

    params: Dict[str, Any] = {
        'Bucket': bucket_name,
        'Key': object_key
    }
    if size is not None:
        params['ContentLength'] = size

    return s3_client.generate_presigned_url(
        'put_object',
        Params=params,
        ExpiresIn=30 * 60 # 30 minutes
    )

def _get_bucket_name_from_uri(bucket_uri: str) -> str:
    if not bucket_uri:
        return ''
    return bucket_uri.split('?')[0].split('/')[2]

def _get_region_name_from_uri(bucket_uri: str) -> str:
    # for example: s3://bucket-name?region=us-west-2
    if not bucket_uri:
        return ''
    default_region = 'auto' # for cloudflare
    a = bucket_uri.split('?')
    if len(a) == 1:
        return default_region
    b = a[1].split('&')
    c = [x for x in b if x.startswith('region=')]
    if len(c) == 0:
        return default_region
    d = c[0].split('=')
    if len(d) != 2:
        return default_region
    return d[1]



import os
import time
from .common import _process_is_alive
from ..common._api_request import _processor_get_api_request


def job_status_monitor(parent_pid: str):
    """
    Monitor job status to see if the job was canceled.
    """
    job_id = os.environ.get('JOB_ID', None)
    if job_id is None:
        raise KeyError('JOB_ID is not set')
    job_private_key = os.environ.get('JOB_PRIVATE_KEY', None)
    if job_private_key is None:
        raise KeyError('JOB_PRIVATE_KEY is not set')
    cancel_out_file = os.environ.get('CANCEL_OUT_FILE', None)
    if cancel_out_file is None:
        raise KeyError('CANCEL_OUT_FILE is not set')

    last_check_timestamp = 0
    overall_timer = time.time()

    while True:
        if not _process_is_alive(parent_pid):
            print(f'Parent process {parent_pid} is no longer alive. Exiting.')
            break

        elapsed_since_check = time.time() - last_check_timestamp
        overall_elapsed = time.time() - overall_timer
        if overall_elapsed < 60:
            interval = 10
        elif overall_elapsed < 60 * 5:
            interval = 30
        elif overall_elapsed < 60 * 20:
            interval = 60
        else:
            interval = 120
        if elapsed_since_check >= interval:
            last_check_timestamp = time.time()
            try:
                status = _get_job_status(job_id=job_id, job_private_key=job_private_key)
                if status != 'running':
                    print(f'Job status is {status}. Canceling.')
                    with open(cancel_out_file, 'w') as f:
                        if isinstance(status, str):
                            f.write(status)
                        else:
                            f.write('0')
                    break
                else:
                    print('Job status is running')
            except: # noqa
                # maybe there was a network error
                print('Error getting job status')

        time.sleep(1)

def _get_job_status(*, job_id: str, job_private_key: str) -> str:
    """Get a job status from the dendro API"""
    url_path = f'/api/processor/jobs/{job_id}/status'
    headers = {
        'job-private-key': job_private_key
    }
    res = _processor_get_api_request(
        url_path=url_path,
        headers=headers
    )
    return res['status']



from typing import List
import queue
from pubnub.pnconfiguration import PNConfiguration
from pubnub.callbacks import SubscribeCallback
from pubnub.pubnub import PubNub

class MySubscribeCallback(SubscribeCallback):
    def __init__(self, message_queue: queue.Queue, compute_resource_id: str):
        self._message_queue = message_queue
        self._compute_resource_id = compute_resource_id
    def message(self, pubnub, message):
        msg = message.message
        if msg.get('computeResourceId', None) == self._compute_resource_id:
            self._message_queue.put(msg)

class PubsubClient:
    def __init__(self, *,
        pubnub_subscribe_key: str,
        pubnub_channel: str,
        pubnub_user: str,
        compute_resource_id: str
    ):
        self._message_queue = queue.Queue()
        pnconfig = PNConfiguration()
        pnconfig.subscribe_key = pubnub_subscribe_key # type: ignore (not sure why we need to type ignore this)
        pnconfig.user_id = pubnub_user
        self._pubnub = PubNub(pnconfig)
        self._listener = MySubscribeCallback(message_queue=self._message_queue, compute_resource_id=compute_resource_id)
        self._pubnub.add_listener(self._listener)
        self._pubnub.subscribe().channels([pubnub_channel]).execute()
    def take_messages(self) -> List[dict]:
        ret = []
        while True:
            try:
                msg = self._message_queue.get(block=False)
                ret.append(msg)
            except queue.Empty:
                break
        return ret
    def close(self):
        self._pubnub.unsubscribe_all()
        self._pubnub.stop()
        self._pubnub.remove_listener(self._listener)
        # unfortunately this doesn't actually kill the thread
        # I submitted a ticket to pubnub about this
        # and they acknowledged that it's a problem
        # but they don't seem to be fixing it



import shutil
import tempfile
from dendro import BaseModel, Field
from dendro.sdk import App, ProcessorBase, InputFile, OutputFile
from dendro.mock import set_use_mock


def test_app():
    set_use_mock(True)

    try:
        class Processor1Context(BaseModel):
            input_file: InputFile = Field(description='Input file')
            output_file: OutputFile = Field(description='Output file')
            text1: str = Field(description='Text 1', default='abc')
            text2: str = Field(description='Text 2')
            text3: str = Field(description='Text 3', default='xyz', json_schema_extra={'options': ['abc', 'xyz']})
            val1: float = Field(description='Value 1', default=1.0)

        class Processor1(ProcessorBase):
            name = 'processor1'
            description = 'This is processor 1'
            label = 'Processor 1'
            tags = ['processor1', 'test']
            attributes = {'test': True}

            @staticmethod
            def run(context: Processor1Context):
                assert context.text1 != ''

        app = App(
            name='test-app',
            description='This is a test app',
            app_image='fake-image'
        )

        app.add_processor(Processor1)

        spec = app.get_spec()
        assert spec['name'] == 'test-app'
    finally:
        set_use_mock(False)

class TemporaryDirectory:
    """A context manager for temporary directories"""
    def __init__(self):
        self._dir = None
    def __enter__(self):
        self._dir = tempfile.mkdtemp()
        return self._dir
    def __exit__(self, exc_type, exc_value, traceback):
        if self._dir:
            shutil.rmtree(self._dir)



import os
import subprocess
from pathlib import Path


def make_app_spec_file_function(app_dir: str, spec_output_file: str):
    # Ensure the directory path is an absolute path
    app_dir_path = Path(app_dir).resolve()

    if spec_output_file is None:
        spec_output_file = str(app_dir_path / 'spec.json')

    script_path = str(app_dir_path / 'main.py')

    # check whether script path is executable
    if not os.access(script_path, os.X_OK):
        raise Exception(f"Script {script_path} is not executable")

    env = os.environ.copy()
    env['SPEC_OUTPUT_FILE'] = spec_output_file
    subprocess.run([script_path], env=env)

    # When we do it the following way, the inspection of type hints in the processor context does not work

    # # Construct the absolute path to main.py in the specified directory
    # main_module_path = app_dir_path / 'main.py'

    # # Check if main.py exists
    # if not main_module_path.exists():
    #     raise FileNotFoundError(f"main.py not found in {app_dir_path}")

    # # Create a module name from the directory path
    # module_name = app_dir_path.name

    # # Use importlib to load the module
    # spec = importlib.util.spec_from_file_location(module_name, str(main_module_path))
    # if spec is None:
    #     raise ImportError(f"Unable to get spec for module {module_name} from {main_module_path}")
    # module = importlib.util.module_from_spec(spec)
    # if spec.loader is None:
    #     raise ImportError(f"Unable to get loader for module {module_name} from {main_module_path}")
    # spec.loader.exec_module(module)

    # # Check if the App class exists in the loaded module
    # if hasattr(module, 'app') and isinstance(getattr(module, 'app'), pr.App):
    #     # Create an instance of the App class
    #     app_instance = module.app

    #     # Call the make_spec_file method
    #     app_instance.make_spec_file(spec_output_file)
    # else:
    #     raise AttributeError("App class not found in main.py")



import json
import aiohttp
import urllib.parse
from ..core.settings import get_settings
from ...mock import using_mock


class PubsubError(Exception):
    pass

async def publish_pubsub_message(*, channel: str, message: dict):
    settings = get_settings()
    # see https://www.pubnub.com/docs/sdks/rest-api/publish-message-to-channel
    sub_key = settings.PUBNUB_SUBSCRIBE_KEY
    pub_key = settings.PUBNUB_PUBLISH_KEY
    uuid = 'dendro'
    # payload is url encoded json
    payload = json.dumps(message)
    payload = urllib.parse.quote(payload)
    url = f"https://ps.pndsn.com/publish/{pub_key}/{sub_key}/0/{channel}/0/{payload}?uuid={uuid}"

    headers = {
        'Accept': 'application/json'
    }

    if using_mock():
        # don't actually publish the message for the mock case
        return True

    # async http get request
    async with aiohttp.ClientSession() as session: # pragma: no cover
        async with session.get(url, headers=headers) as resp:
            if resp.status != 200:
                raise PubsubError(f"Error publishing to pubsub: {resp.status} {resp.text}")
            return True



from typing import TYPE_CHECKING, List
import os
import time
import subprocess

from ..mock import using_mock
from ..common.dendro_types import DendroJob, DendroJobRequiredResources

if TYPE_CHECKING:
    from .JobManager import JobManager


max_num_simultaneous_slurm_groups = 5 # TODO: make this configurable

# This will determine how many cpu jobs get grouped together in a single slurm job
num_cpus_per_slurm_node = 100 # TODO: make this configurable

class SlurmJobHandler:
    def __init__(self, *, job_manager: 'JobManager'):
        self._job_manager = job_manager
        self._last_time_we_got_a_new_job = 0
        self._jobs: List[DendroJob] = []
        self._job_ids_attempted_to_start_or_fail = set() # we don't want to attempt to start the same job multiple times
        self._last_disk_cleanup_time = 0
        # We don't want to store any state in memory!
        # so if we restart the compute resource, we don't want that to interrupt the process of knowing when
        # to start new job groups.
    def handle_jobs(self, jobs: List[DendroJob]):
        current_job_ids = set([job.jobId for job in self._jobs])
        for job in jobs:
            if job.jobId not in current_job_ids:
                self._last_time_we_got_a_new_job = time.time()
        self._jobs = jobs
    def do_work(self):
        if time.time() > self._last_disk_cleanup_time + 300:
            _clean_up_old_files()
            self._last_disk_cleanup_time = time.time()

        time_scale_factor = 1 if not using_mock() else 10000

        elapsed_since_got_new_job = time.time() - self._last_time_we_got_a_new_job
        if elapsed_since_got_new_job < 5 / time_scale_factor:
            # wait because we might still have jobs coming in and we need to start them in groups
            return
        if len(self._jobs) == 0:
            # we have no jobs
            return
        pending_jobs = [job for job in self._jobs if job.status == 'pending' and job.jobId not in self._job_ids_attempted_to_start_or_fail]
        pending_job_groups = _split_jobs_into_groups(pending_jobs)

        num_running_groups = self._determine_num_running_groups()

        for job_group in pending_job_groups:
            if num_running_groups >= max_num_simultaneous_slurm_groups:
                break
            self._start_job_group(job_group)
            num_running_groups += 1
    def _start_job_group(self, job_group: 'PendingJobGroup'):
        # a job group is a collection of jobs that should ideally fit in a single node on the cluster
        jobs = job_group._jobs
        required_resources = job_group._required_resources
        for job in jobs:
            self._job_ids_attempted_to_start_or_fail.add(job.jobId)

        if not os.path.exists('slurm_scripts'):
            os.mkdir('slurm_scripts')
        if not os.path.exists('slurm_group_assignments'):
            os.mkdir('slurm_group_assignments')

        num_cpus_per_job = required_resources.numCpus
        num_gpus_per_job = required_resources.numGpus
        memory_gb_per_job = required_resources.memoryGb
        timeout_sec_per_job = required_resources.timeSec

        if num_gpus_per_job > 1:
            self._fail_jobs(jobs, 'numGpus > 1 is not supported (SlurmBatchHandler)')
            return

        random_str = os.urandom(16).hex()
        timestamp = _create_nice_timestamp()

        slurm_group_id = f'{timestamp}_{random_str}'
        slurm_script_fname = f'slurm_scripts/slurm_batch_{slurm_group_id}.sh'
        script_has_at_least_one_job = False # important to do this so we don't ever run an empty script
        with open(slurm_script_fname, 'w', encoding='utf8') as f:
            f.write('#!/bin/bash\n')
            f.write('\n')
            f.write('set -e\n')
            f.write('\n')
            for ii, job in enumerate(jobs):
                # doesn't actually start the job - just returns the shell command - but sets the job status to 'starting'
                # so it's important to run this slurm script right away so that the job doesn't get caught in a starting state
                # forever, for example if the compute resource is restarted
                cmd = self._job_manager._start_job(job, run_process=False, return_shell_command=True)
                if cmd:
                    if not using_mock():
                        # this is how we make sure one and only one job is run in each process
                        f.write(f'if [ "$SLURM_PROCID" == "{ii}" ]; then\n')
                        f.write(f'    {cmd}\n')
                        f.write('fi\n')
                        f.write('\n')
                    else:
                        f.write(f'{cmd}\n')
                        f.write('\n')
                    script_has_at_least_one_job = True

                    # we need to record which slurm group this job went to so we can
                    # later know how many running groups there are
                    with open(f'slurm_group_assignments/{job.jobId}', 'w', encoding='utf8') as f2:
                        f2.write(slurm_group_id)
            f.write('\n')
        if script_has_at_least_one_job:
            # run the slurm script with srun
            # slurm_cpus_per_task = self._slurm_opts.cpusPerTask
            # slurm_partition = self._slurm_opts.partition
            # slurm_time = self._slurm_opts.time
            # slurm_other_opts = self._slurm_opts.otherOpts

            if num_gpus_per_job > 0:
                slurm_partition = os.environ.get('SLURM_GPU_PARTITION', None)
                if not slurm_partition:
                    raise Exception('SLURM_GPU_PARTITION environment variable must be set')
            else:
                slurm_partition = os.environ.get('SLURM_PARTITION', None)
                if not slurm_partition:
                    raise Exception('SLURM_PARTITION environment variable must be set')

            slurm_time = _format_time_for_slurm(timeout_sec_per_job)
            if slurm_time is None:
                self._fail_jobs(jobs, f'Invalid timeout_sec_per_job: {timeout_sec_per_job}')
                return

            oo = [] # the slurm options
            oo.append(f'-n {len(jobs)}') # number of tasks
            oo.append(f'--cpus-per-task={num_cpus_per_job}')
            oo.append(f'--partition={slurm_partition}')
            if num_gpus_per_job > 0:
                oo.append(f'--gpus-per-task={num_gpus_per_job}')
            oo.append(f'--time={slurm_time}')

            # provide memory per cpu
            mem_per_cpu = int(memory_gb_per_job * 1024 / num_cpus_per_job)
            oo.append(f'--mem-per-cpu={mem_per_cpu}')

            slurm_opts_str = ' '.join(oo)
            if not using_mock():
                cmd = f'srun {slurm_opts_str} bash {slurm_script_fname}'
            else:
                cmd = f'bash {slurm_script_fname}'

            print(f'Running slurm batch: {cmd}')
            subprocess.Popen(
                cmd.split(' '),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                start_new_session=True
            )
    def _fail_jobs(self, jobs: List[DendroJob], reason: str):
        for job in jobs:
            self._job_manager._fail_job(job, reason)
            self._job_ids_attempted_to_start_or_fail.add(job.jobId)
    def _determine_num_running_groups(self) -> int:
        # we need to know how many slurm groups are currently running
        # so we can limit the number of groups we start at once
        all_group_ids = []
        for job in self._jobs:
            if job.status == 'starting' or job.status == 'running':
                if os.path.exists(f'slurm_group_assignments/{job.jobId}'):
                    with open(f'slurm_group_assignments/{job.jobId}', 'r', encoding='utf8') as f:
                        group_id = f.read().strip()
                        all_group_ids.append(group_id)
        return len(set(all_group_ids))

class PendingJobGroup:
    def __init__(self, jobs: List[DendroJob], required_resources: DendroJobRequiredResources):
        self._jobs = jobs
        self._required_resources = required_resources

def _split_jobs_into_groups(jobs: List[DendroJob]) -> List[PendingJobGroup]:
    jobs = [job for job in jobs if job.requiredResources is not None]
    ret: List[PendingJobGroup] = []
    all_batch_ids = set([job.batchId for job in jobs if job.batchId is not None])
    for batch_id in all_batch_ids:
        jobs_in_batch = [job for job in jobs if job.batchId == batch_id]
        assert len(jobs_in_batch) > 0
        job0 = jobs_in_batch[0]
        # we assume that all jobs here have the same required resources
        required_resources = job0.requiredResources
        assert required_resources is not None
        if required_resources.numGpus > 0:
            # every gpu job goes to their own group
            for job in jobs_in_batch:
                ret.append(PendingJobGroup([job], required_resources))
        else:
            # split into groups based on how many can fit on a single node, determined by numCpus
            # TODO: take into account memory requirements
            num_tasks_per_node = num_cpus_per_slurm_node // required_resources.numCpus
            if num_tasks_per_node == 0:
                num_tasks_per_node = 1
            for i in range(0, len(jobs_in_batch), num_tasks_per_node):
                ret.append(PendingJobGroup(jobs_in_batch[i:i + num_tasks_per_node], required_resources))
    for job in jobs:
        if job.batchId is None:
            assert job.requiredResources is not None
            ret.append(PendingJobGroup([job], job.requiredResources))
    return ret

def _format_time_for_slurm(timeout_sec: float):
    # format is d-hh:mm:ss
    timeout_sec = int(timeout_sec)
    if timeout_sec <= 0:
        return None
    d = timeout_sec // (24 * 3600)
    timeout_sec = timeout_sec % (24 * 3600)
    h = timeout_sec // 3600
    timeout_sec %= 3600
    m = timeout_sec // 60
    timeout_sec %= 60
    s = timeout_sec
    if d > 0:
        return f'{d}-{h:02d}:{m:02d}:{s:02d}'
    else:
        return f'{h:02d}:{m:02d}:{s:02d}'

def _create_nice_timestamp():
    return time.strftime('%Y-%m-%d-%H-%M-%S', time.localtime())

def _clean_up_old_files():
    # delete old slurm scripts
    if os.path.exists('slurm_scripts'):
        for fname in os.listdir('slurm_scripts'):
            if fname.startswith('slurm_batch_'):
                full_fname = os.path.join('slurm_scripts', fname)
                if os.path.isfile(full_fname):
                    elapsed_since_modified = time.time() - os.path.getmtime(full_fname)
                    if elapsed_since_modified > 60 * 60 * 48:
                        os.remove(full_fname)
    # delete old slurm group assignments
    if os.path.exists('slurm_group_assignments'):
        for fname in os.listdir('slurm_group_assignments'):
            full_fname = os.path.join('slurm_group_assignments', fname)
            if os.path.isfile(full_fname):
                elapsed_since_modified = time.time() - os.path.getmtime(full_fname)
                if elapsed_since_modified > 60 * 60 * 48:
                    os.remove(full_fname)



import asyncio
from ..core.settings import get_settings
from .MockMongoClient import MockMongoClient
from ...mock import using_mock

_globals = {"mock_mongo_client": None}


# pyright: reportGeneralTypeIssues=false
def _get_mongo_client():
    # We want one async mongo client per event loop
    loop = asyncio.get_event_loop()
    if hasattr(loop, "_mongo_client"):
        return loop._mongo_client  # type: ignore

    mongo_uri = get_settings().MONGO_URI

    # If we're using a mock client, return it
    if using_mock():
        client = _globals["mock_mongo_client"]  # type: ignore
        if client is None:
            client = MockMongoClient()
            _globals["mock_mongo_client"] = client  # type: ignore
    else:  # pragma: no cover
        # Otherwise, create a new client
        assert mongo_uri is not None, "MONGO_URI environment variable not set"
        from motor.motor_asyncio import AsyncIOMotorClient

        client = AsyncIOMotorClient(mongo_uri)

    # Store the client on the event loop
    setattr(loop, "_mongo_client", client)

    return client


def _clear_mock_mongo_databases():
    client: MockMongoClient = _globals["mock_mongo_client"]  # type: ignore
    if client is not None:
        client.clear_databases()



import os
import time
import traceback
import json
import psutil
from .common import _get_upload_url, _process_is_alive


def resource_utilization_monitor(parent_pid: str):
    """
    Monitor resource utilization of a job.
    """
    job_id = os.environ.get('JOB_ID', None)
    if job_id is None:
        raise KeyError('JOB_ID is not set')
    job_private_key = os.environ.get('JOB_PRIVATE_KEY', None)
    if job_private_key is None:
        raise KeyError('JOB_PRIVATE_KEY is not set')
    last_report_timestamp = 0
    last_upload_timestamp = 0
    overall_timer = time.time()
    all_lines = []

    while True:
        if not _process_is_alive(parent_pid):
            print(f'Parent process {parent_pid} is no longer alive. Doing final upload.')
            do_upload(
                all_lines=all_lines,
                job_id=job_id,
                job_private_key=job_private_key
            )
            print('Exiting.')
            break

        elapsed_since_report = time.time() - last_report_timestamp
        if elapsed_since_report >= 10:
            last_report_timestamp = time.time()
            cpu_percent = psutil.cpu_percent()
            virtual_memory = psutil.virtual_memory()
            disk_io_counters = psutil.disk_io_counters()
            net_io_counters = psutil.net_io_counters(pernic=False, nowrap=True)
            gpu_loads = _get_gpu_loads()
            log_record = {
                'timestamp': time.time(),
                'cpu': {
                    'percent': cpu_percent
                },
                'virtual_memory': {
                    'total': virtual_memory.total,
                    'available': virtual_memory.available,
                    'percent': virtual_memory.percent,
                    'used': virtual_memory.used,
                    'free': virtual_memory.free,
                    'active': virtual_memory.active,
                    'inactive': virtual_memory.inactive,
                    'buffers': virtual_memory.buffers,
                    'cached': virtual_memory.cached,
                    'shared': virtual_memory.shared,
                    'slab': virtual_memory.slab
                },
                'disk_io_counters': {
                    'read_count': disk_io_counters.read_count,
                    'write_count': disk_io_counters.write_count,
                    'read_bytes': disk_io_counters.read_bytes,
                    'write_bytes': disk_io_counters.write_bytes,
                    'read_time': disk_io_counters.read_time,
                    'write_time': disk_io_counters.write_time
                } if disk_io_counters else None,
                'net_io_counters': {
                    'bytes_sent': net_io_counters.bytes_sent,
                    'bytes_recv': net_io_counters.bytes_recv,
                    'packets_sent': net_io_counters.packets_sent,
                    'packets_recv': net_io_counters.packets_recv,
                    'errin': net_io_counters.errin,
                    'errout': net_io_counters.errout,
                    'dropin': net_io_counters.dropin,
                    'dropout': net_io_counters.dropout
                },
                'gpu': {
                    'loads': gpu_loads
                } if gpu_loads else None
            }
            line = (json.dumps(log_record) + '\n').encode('utf-8') # important to encode this string
            all_lines.append(line)
            elapsed_since_upload = time.time() - last_upload_timestamp
            overall_elapsed = time.time() - overall_timer
            if overall_elapsed < 60:
                interval = 10
            elif overall_elapsed < 60 * 5:
                interval = 30
            elif overall_elapsed < 60 * 20:
                interval = 60
            else:
                interval = 120
            if elapsed_since_upload >= interval:
                last_upload_timestamp = time.time()
                do_upload(
                    all_lines=all_lines,
                    job_id=job_id,
                    job_private_key=job_private_key
                )

        time.sleep(1)

def _get_gpu_loads():
    """This is a helper function for _resource_utilization_log_reader"""
    try:
        import GPUtil
        gpus = GPUtil.getGPUs()
        return [gpu.load for gpu in gpus]
    except: # noqa
        return None

def do_upload(*, all_lines, job_id, job_private_key):
    import requests
    print(f'Uploading {len(all_lines)} lines of resource utilization data')
    text_to_upload = b'\n'.join(all_lines)
    try:
        resource_utilization_log_upload_url = _get_upload_url(job_id=job_id, job_private_key=job_private_key, output_name='_resource_utilization_log')
        r = requests.put(resource_utilization_log_upload_url, data=text_to_upload, timeout=5)
        if r.status_code != 200:
            raise Exception(f'Error uploading resource utilization log: {r.status_code} {r.text}')
    except: # noqa
        print('Error uploading resource utilization log')
        traceback.print_exc()



import os
import json
from fastapi import APIRouter, Header
from .... import BaseModel
from ....common.dendro_types import ComputeResourceUserUsage
from ._authenticate_gui_request import _authenticate_gui_request
from ..common import api_route_wrapper
from ...services.gui.get_compute_resource_user_usage import get_compute_resource_user_usage


router = APIRouter()

# get usage
class GetUsageResponse(BaseModel):
    usage: ComputeResourceUserUsage
    success: bool

@router.get("/compute_resource/{compute_resource_id}/user/{user_id}")
@api_route_wrapper
async def get_usage(compute_resource_id, user_id, github_access_token: str = Header(...)):
    # url decode the %7C in the user_id
    user_id = user_id.replace('%7C', '|')

    # authenticate the request
    auth_user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert auth_user_id

    if user_id != auth_user_id:
        ADMIN_USER_IDS_JSON = os.getenv('ADMIN_USER_IDS', '[]')
        ADMIN_USER_IDS = json.loads(ADMIN_USER_IDS_JSON)

        if user_id not in ADMIN_USER_IDS:
            raise Exception(f'User is not admin and cannot view usage for other users ({user_id} != {auth_user_id})')

    # get usage
    usage = await get_compute_resource_user_usage(compute_resource_id=compute_resource_id, user_id=user_id)

    return GetUsageResponse(usage=usage, success=True)



from typing import Any, List, Union, Literal
import os
from pydantic import BaseModel
from .Project import Project
from ..common.dendro_types import CreateJobRequest, CreateJobResponse, CreateJobRequestInputFile, CreateJobRequestOutputFile, CreateJobRequestInputParameter, ComputeResourceSpecProcessor, DendroJob, DendroJobRequiredResources
from ..common._api_request import _client_post_api_request


class SubmitJobInputFile(BaseModel):
    name: str
    file_name: str
    is_folder: Union[bool, None] = None

class SubmitJobOutputFile(BaseModel):
    name: str
    file_name: str
    is_folder: Union[bool, None] = None

class SubmitJobParameter(BaseModel):
    name: str
    value: Any

def submit_job(*,
    project: Project,
    processor_name: str,
    input_files: List[SubmitJobInputFile],
    output_files: List[SubmitJobOutputFile],
    parameters: List[SubmitJobParameter],
    batch_id: Union[str, None] = None,
    rerun_policy: str = 'never', # always | never | if_failed
    required_resources: DendroJobRequiredResources,
    run_method: Literal['local', 'aws_batch', 'slurm']
):
    """Submit a job to the Dendro compute service.

    Args:
        project (Project): The project to submit the job to.
        processor_name (str): The name of the processor to use.
        input_files (List[SubmitJobInputFile]): The input files for the job.
        output_files (List[SubmitJobOutputFile]): The output files for the job.
        parameters (List[SubmitJobParameter]): The input parameters for the job.
        batch_id (Union[str, None], optional): The batch ID to use. Defaults to None.
        rerun_policy (str, optional): The rerun policy to use. One of: 'always', 'never', 'if_failed'. Defaults to 'never'.

    Returns:
        str: The job ID.
    """
    if rerun_policy not in ['always', 'never', 'if_failed']:
        raise ValueError(f'Invalid rerun policy: {rerun_policy}')
    compute_resource = project._compute_resource
    if not compute_resource:
        raise KeyError(f'No compute resource found for project {project._name} ({project._project_id})')
    processor_spec = None
    if compute_resource.spec is None:
        raise KeyError(f'No spec found for compute resource {compute_resource.name} for project {project._name} ({project._project_id})')
    for app in compute_resource.spec.apps:
        for processor in app.processors:
            if processor.name == processor_name:
                processor_spec = processor
                break
        if processor_spec:
            break
    if not processor_spec:
        raise KeyError(f'No processor with name {processor_name} found in compute resource {compute_resource.name} for project {project._name} ({project._project_id})')
    default_parameters = _create_default_parameters(processor_spec, parameters)
    parameters = parameters + default_parameters
    _check_consistency_with_processor_spec(
        processor_spec=processor_spec,
        processor_name=processor_name,
        input_files=input_files,
        output_files=output_files,
        parameters=parameters
    )
    request_input_files = [
        CreateJobRequestInputFile(
            name=x.name,
            fileName=x.file_name,
            isFolder=x.is_folder
        )
        for x in input_files
    ]
    request_output_files = [
        CreateJobRequestOutputFile(
            name=x.name,
            fileName=x.file_name,
            isFolder=x.is_folder
        )
        for x in output_files
    ]
    request_parameters = [
        CreateJobRequestInputParameter(
            name=x.name,
            value=x.value
        )
        for x in parameters
    ]
    matching_job = None
    for job in project._jobs:
        if _job_matches(
            job=job,
            processor_name=processor_name,
            input_files=input_files,
            output_files=output_files,
            parameters=parameters
        ):
            matching_job = job
            break
    if matching_job:
        rerun = True
        if rerun_policy == 'never':
            rerun = False
        elif rerun_policy == 'if_failed':
            if matching_job.status == 'failed':
                rerun = True
            else:
                rerun = False
        elif rerun_policy == 'always':
            rerun = True
        else:
            raise ValueError(f'Invalid rerun policy: {rerun_policy}')
        if not rerun:
            print(f'Skipping job submission because a matching job was found: {matching_job.jobId}')
            return matching_job.jobId
    req = CreateJobRequest(
        projectId=project._project_id,
        processorName=processor_name,
        inputFiles=request_input_files,
        outputFiles=request_output_files,
        inputParameters=request_parameters,
        processorSpec=processor_spec,
        batchId=batch_id,
        dandiApiKey=os.environ.get('DANDI_API_KEY', None),
        requiredResources=required_resources,
        runMethod=run_method
    )

    dendro_api_key = os.environ.get('DENDRO_API_KEY', None)
    if not dendro_api_key:
        raise ValueError('DENDRO_API_KEY environment variable is not set')

    url_path = '/api/client/jobs'
    resp = _client_post_api_request(
        url_path=url_path,
        data=_model_dump(req),
        dendro_api_key=dendro_api_key
    )
    resp = CreateJobResponse(**resp)
    print(f'Submitted job: {resp.jobId}')
    return resp.jobId

def _create_default_parameters(
    processor_spec: ComputeResourceSpecProcessor,
    parameters: List[SubmitJobParameter]
) -> List[SubmitJobParameter]:
    default_parameters = []
    for pp in processor_spec.parameters:
        if not any(x.name == pp.name for x in parameters):
            # need to wait for the model to be updated to include the "required" flag to do this
            # if pp.required:
            #     raise KeyError(f"Required parameter not found: {pp.name}")
            default_parameters.append(SubmitJobParameter(name=pp.name, value=pp.default))
    return default_parameters

def _check_consistency_with_processor_spec(
    processor_spec: ComputeResourceSpecProcessor,
    processor_name: str,
    input_files: List[SubmitJobInputFile],
    output_files: List[SubmitJobOutputFile],
    parameters: List[SubmitJobParameter]
):
    # check that the processor name matches
    if processor_spec.name != processor_name:
        raise KeyError(f'Processor name mismatch: {processor_spec.name} != {processor_name}')
    # check that the input files are consistent with the spec
    for input_file in input_files:
        if not input_file.is_folder:
            pp = next((x for x in processor_spec.inputs if x.name == input_file.name), None)
            if not pp:
                raise KeyError(f"Processor input not found: {input_file.name}")
        else:
            pp = next((x for x in (processor_spec.inputFolders or []) if x.name == input_file.name), None)
            if not pp:
                raise KeyError(f"Processor input folder not found: {input_file.name}")

    # check that no input files are missing
    for input in processor_spec.inputs:
        if not any(x.name == input.name and not x.is_folder for x in input_files):
            raise KeyError(f"Required input not found: {input.name}")
    # check that no input folders are missing
    for input_folder in processor_spec.inputFolders or []:
        if not any(x.name == input_folder.name and x.is_folder for x in input_files):
            raise KeyError(f"Required input folder not found: {input_folder.name}")
    # check that the output files are consistent with the spec
    for output_file in output_files:
        if not output_file.is_folder:
            pp = next((x for x in processor_spec.outputs if x.name == output_file.name), None)
            if not pp:
                raise KeyError(f"Processor output not found: {output_file.name}")
        else:
            pp = next((x for x in (processor_spec.outputFolders or []) if x.name == output_file.name), None)
            if not pp:
                raise KeyError(f"Processor output folder not found: {output_file.name}")
    # check that no output files are missing
    for output in processor_spec.outputs:
        if not any(x.name == output.name and not x.is_folder for x in output_files):
            raise KeyError(f"Required output not found: {output.name}")
    # check that no output folders are missing
    for output_folder in processor_spec.outputFolders or []:
        if not any(x.name == output_folder.name and x.is_folder for x in output_files):
            raise KeyError(f"Required output folder not found: {output_folder.name}")
    # check that the input parameters are consistent with the spec
    for input_parameter in parameters:
        pp = next((x for x in processor_spec.parameters if x.name == input_parameter.name), None)
        if not pp:
            raise KeyError(f"Processor parameter not found: {input_parameter.name}")

    # For this we need to wait for the model to be updated to include the "required" flag
    # # check that no required parameters are missing
    # for pp in processor_spec.parameters:
    #     if pp.required:
    #         input_parameter = next((x for x in parameters if x.name == pp.name), None)
    #         if not input_parameter:
    #             raise KeyError(f"Required parameter not found: {pp.name}")

def _model_dump(model, exclude_none=False):
    # handle both pydantic v1 and v2
    if hasattr(model, 'model_dump'):
        return model.model_dump(exclude_none=exclude_none)
    else:
        return model.dict(exclude_none=exclude_none)

def _job_matches(*,
    job: DendroJob,
    processor_name: str,
    input_files: List[SubmitJobInputFile],
    output_files: List[SubmitJobOutputFile],
    parameters: List[SubmitJobParameter]
):
    if job.processorName != processor_name:
        return False

    # input files
    if len(job.inputFiles) != len(input_files):
        return False
    for input_file in input_files:
        x = next((x for x in job.inputFiles if x.name == input_file.name), None)
        if not x:
            return False
        if x.fileName != input_file.file_name:
            return False
        if x.isFolder != input_file.is_folder:
            return False

    # output files
    if len(job.outputFiles) != len(output_files):
        return False
    for output_file in output_files:
        x = next((x for x in job.outputFiles if x.name == output_file.name), None)
        if not x:
            return False
        if x.fileName != output_file.file_name:
            return False
        if x.isFolder != output_file.is_folder:
            return False

    # input parameters
    if len(job.inputParameters) != len(parameters):
        return False
    for input_parameter in parameters:
        x = next((x for x in job.inputParameters if x.name == input_parameter.name), None)
        if not x:
            return False
        if x.value != input_parameter.value:
            return False
    return True



from typing import Optional
import os
import time
import click
from .compute_resource.register_compute_resource import register_compute_resource as register_compute_resource_function
from .compute_resource.start_compute_resource import start_compute_resource as start_compute_resource_function
from .sdk._make_spec_file import make_app_spec_file_function
from .sdk._test_app_processor import test_app_processor_function

# ------------------------------------------------------------
# Compute resource cli
# ------------------------------------------------------------
@click.command(help='Initialize a compute resource controller in the current directory')
@click.option('--compute-resource-id', default=None, help='Compute resource ID')
@click.option('--compute-resource-private-key', default=None, help='Compute resource private key')
def register_compute_resource(compute_resource_id: str, compute_resource_private_key: str):
    register_compute_resource_function(dir='.', compute_resource_id=compute_resource_id, compute_resource_private_key=compute_resource_private_key)


@click.command(help="Start the compute resource controller in the current directory")
@click.option('--timeout', default=None, type=float, help='Timeout (sec) for the daemon - used for testing')
def start_compute_resource(timeout: Optional[float] = None):
    start_compute_resource_function(dir='.', timeout=timeout)
    # unfortunately, the pubnub background thread cannot be terminated, so we have to force-exit the process
    # I opened a ticket with pubnub about this, and they acknowledged that it's a problem, but they don't seem to be fixing it
    os._exit(0)


# ------------------------------------------------------------
# App cli
# ------------------------------------------------------------
@click.command(help='Make an app spec file')
@click.option('--app-dir', default='.', help='Path to the app directory')
@click.option('--spec-output-file', default=None, help='Output file for the spec')
def make_app_spec_file(app_dir: str, spec_output_file: str):
    make_app_spec_file_function(app_dir=app_dir, spec_output_file=spec_output_file)

# ------------------------------------------------------------
# Mock job cli
# ------------------------------------------------------------
@click.command(help='Run a mock job (used during testing)')
def run_mock_job():
    print('Running mock job')
    time.sleep(0.001) # don't pause too long
    print('Mock job completed')


# ------------------------------------------------------------
# Test app processor
# ------------------------------------------------------------
@click.command(help='Run an app processor (used for testing)')
@click.option('--app-dir', default='.', help='Path to the app directory')
@click.option('--processor', required=True, help='Name of the processor to run')
@click.option('--context', required=True, help='Path to the context file (.json or .yaml)')
def test_app_processor(app_dir: str, processor: str, context: str):
    test_app_processor_function(app_dir=app_dir, processor=processor, context=context)

# ------------------------------------------------------------
# Internal job monitoring process
# ------------------------------------------------------------
@click.command(help='Internal job monitoring process')
@click.argument('monitor_type', type=click.Choice(['resource_utilization', 'console_output', 'job_status']))
@click.option('--parent-pid', required=True, help='Parent PID')
def internal_job_monitor(monitor_type: str, parent_pid: str):
    if monitor_type == 'resource_utilization':
        from .internal_job_monitoring.resource_utilization_monitor import resource_utilization_monitor
        resource_utilization_monitor(parent_pid=parent_pid)
    elif monitor_type == 'console_output':
        from .internal_job_monitoring.console_output_monitor import console_output_monitor
        console_output_monitor(parent_pid=parent_pid)
    elif monitor_type == 'job_status':
        from .internal_job_monitoring.job_status_monitor import job_status_monitor
        job_status_monitor(parent_pid=parent_pid)
    else:
        raise Exception(f'Unrecognized monitor_type: {monitor_type}')

# ------------------------------------------------------------
# Main cli
# ------------------------------------------------------------
@click.group(help="dendro command line interface")
def main():
    pass

main.add_command(register_compute_resource)
main.add_command(start_compute_resource)
main.add_command(make_app_spec_file)
main.add_command(test_app_processor)
main.add_command(run_mock_job)
main.add_command(internal_job_monitor)



from typing import List, Union
import os
import time
import requests
from ..common.dendro_types import DendroComputeResource, DendroProject, DendroFile, DendroJob
from ..common._api_request import _client_get_api_request


class ProjectException(Exception):
    pass

class Project:
    def __init__(self, *,
        project_data: DendroProject,
        files_data: List[DendroFile],
        jobs_data: List[DendroJob],
        compute_resource: Union[DendroComputeResource, None]
    ) -> None:
        self._project_id = project_data.projectId
        self._name = project_data.name
        self._description = project_data.description
        self._compute_resource_id = project_data.computeResourceId
        self._owner_id = project_data.ownerId
        self._timestamp_created = project_data.timestampCreated
        self._timestamp_modified = project_data.timestampModified

        self._files = [
            ProjectFile(f)
            for f in files_data
        ]

        self._jobs = [
            j # DendroJob
            for j in jobs_data
        ]

        self._compute_resource = compute_resource
    def get_file(self, file_name: str) -> Union['ProjectFile', None]:
        for f in self._files:
            if f._file_data.fileName == file_name:
                return f
        return None
    def get_folder(self, path: str) -> 'ProjectFolder':
        return ProjectFolder(self, path)

class ProjectFile:
    def __init__(self, file_data: DendroFile) -> None:
        self._file_data = file_data
        self._resolved_url: Union[str, None] = None
        self._resolved_url_timestamp: Union[float, None] = None
    def get_url(self) -> str:
        a = self._file_data.content
        if not a.startswith('url:'):
            raise ProjectException(f'Unexpected content for file {self._file_data.fileName}: {a}')
        url = a[len('url:'):]
        resolved_url = self._get_resolved_url(url=url)
        return resolved_url
    @property
    def file_name(self) -> str:
        return self._file_data.fileName
    def _get_resolved_url(self, url: str) -> str:
        if self._resolved_url is not None and self._resolved_url_timestamp is not None:
            if time.time() - self._resolved_url_timestamp < 120:
                return self._resolved_url
        resolve_with_dandi_api_key = None
        if url.startswith('https://api.dandiarchive.org/api/'):
            dandi_api_key = os.environ.get('DANDI_API_KEY', None)
            if dandi_api_key is not None:
                resolve_with_dandi_api_key = dandi_api_key
        elif url.startswith('https://api-staging.dandiarchive.org/'):
            dandi_api_key = os.environ.get('DANDI_STAGING_API_KEY', None)
            if dandi_api_key is not None:
                resolve_with_dandi_api_key = dandi_api_key
        if resolve_with_dandi_api_key is not None:
            url0 = _resolve_dandi_url(url=url, dandi_api_key=resolve_with_dandi_api_key)
        else:
            url0 = url
        self._resolved_url = url0
        self._resolved_url_timestamp = time.time()
        return self._resolved_url

class ProjectJob:
    def __init__(self, job_data: DendroJob) -> None:
        self._job_data = job_data

class ProjectFolder:
    def __init__(self, project: Project, path: str) -> None:
        self._project = project
        self._path = path
    @property
    def path(self) -> str:
        return self._path
    def get_files(self) -> List[ProjectFile]:
        ret: List[ProjectFile] = []
        for f in self._project._files:
            a = f._file_data.fileName.split('/')
            ok = False
            if len(a) <= 1:
                ok = (self._path == '')
            else:
                ok = '/'.join(a[:-1]) == self._path
            if ok:
                ret.append(f)
        return ret
    def get_folders(self) -> List['ProjectFolder']:
        folder_paths = set()
        for f in self._project._files:
            a = f._file_data.fileName.split('/')
            if len(a) <= 1:
                continue
            folder_path = '/'.join(a[:-1])
            if folder_path.startswith(self._path + '/'):
                x = folder_path[len(self._path) + 1:]
                if '/' in x:
                    x = x[:x.index('/')]
                folder_paths.add(self._path + '/' + x)
        sorted_folder_paths = sorted(list(folder_paths))
        return [
            ProjectFolder(self._project, p)
            for p in sorted_folder_paths
        ]

def load_project(project_id: str) -> Project:
    url_path = f'/api/client/projects/{project_id}'
    project_resp = _client_get_api_request(url_path=url_path)
    project: DendroProject = DendroProject(**project_resp['project'])

    url_path = f'/api/client/projects/{project_id}/files'
    files_resp = _client_get_api_request(url_path=url_path)
    files: List[DendroFile] = [DendroFile(**f) for f in files_resp['files']]

    url_path = f'/api/client/projects/{project_id}/jobs'
    jobs_resp = _client_get_api_request(url_path=url_path)
    jobs: List[DendroJob] = [DendroJob(**j) for j in jobs_resp['jobs']]

    if project.computeResourceId:
        url_path = f'/api/client/compute_resources/{project.computeResourceId}'
        compute_resource_resp = _client_get_api_request(url_path=url_path)
        compute_resource = DendroComputeResource(**compute_resource_resp['computeResource'])
    else:
        compute_resource = None

    return Project(project_data=project, files_data=files, jobs_data=jobs, compute_resource=compute_resource)

def _resolve_dandi_url(*, url: str, dandi_api_key: str):
    headers = {
        'Authorization': f'token {dandi_api_key}'
    }
    # do it synchronously here
    resp = requests.head(url, allow_redirects=True, headers=headers)
    return str(resp.url)



from typing import Optional
import os

# Note: BaseSettings is no longer a part of pydantic package, and I didn't want to add a dependency on pydantic-settings
# So I'm using BaseModel instead


class Settings:
    def __init__(self) -> None:
        # General app config
        self.MONGO_URI: Optional[str] = os.environ.get("MONGO_URI")

        self.PUBNUB_SUBSCRIBE_KEY: Optional[str] = os.environ.get("VITE_PUBNUB_SUBSCRIBE_KEY")
        self.PUBNUB_PUBLISH_KEY: Optional[str] = os.environ.get("PUBNUB_PUBLISH_KEY")

        self.GITHUB_CLIENT_ID: Optional[str] = os.environ.get("VITE_GITHUB_CLIENT_ID")
        self.GITHUB_CLIENT_SECRET: Optional[str] = os.environ.get("GITHUB_CLIENT_SECRET")

        self.DEFAULT_COMPUTE_RESOURCE_ID: Optional[str] = os.environ.get("VITE_DEFAULT_COMPUTE_RESOURCE_ID")

        self.OUTPUT_BUCKET_URI: Optional[str] = os.environ.get("OUTPUT_BUCKET_URI")
        self.OUTPUT_BUCKET_CREDENTIALS: Optional[str] = os.environ.get("OUTPUT_BUCKET_CREDENTIALS")
        self.OUTPUT_BUCKET_BASE_URL: Optional[str] = os.environ.get("OUTPUT_BUCKET_BASE_URL")

def get_settings():
    return Settings()



import os
from typing import List, Union, Literal

from ..sdk.App import App
from ..common.dendro_types import DendroComputeResourceApp
from ..common._api_request import _compute_resource_get_api_request, _compute_resource_put_api_request


class AppManager:
    def __init__(self, *,
                 compute_resource_id: str,
                 compute_resource_private_key: str,
                 available_job_run_methods: List[Literal['local', 'aws_batch', 'slurm']]
                ):
        self._compute_resource_id = compute_resource_id
        self._compute_resource_private_key = compute_resource_private_key
        self._available_job_run_methods = available_job_run_methods

        self._compute_resource_apps: List[DendroComputeResourceApp] = []
        self._apps: List[App] = []

    def update_apps(self):
        url_path = f'/api/compute_resource/compute_resources/{self._compute_resource_id}/apps'
        resp = _compute_resource_get_api_request(
            url_path=url_path,
            compute_resource_id=self._compute_resource_id,
            compute_resource_private_key=self._compute_resource_private_key
        )

        # It would be nice to do it this way, but we can't because we don't want to import stuff from the api here
        # from ..api_helpers.routers.compute_resource.router import GetAppsResponse
        # resp = GetAppsResponse(**resp)
        # compute_resource_apps = resp.apps

        # instead:
        compute_resource_apps_from_response = resp['apps']
        compute_resource_apps_from_response = [DendroComputeResourceApp(**app) for app in compute_resource_apps_from_response]
        something_changed = False
        for app in self._apps:
            matching_cr_app = next((cr_app for cr_app in compute_resource_apps_from_response if _app_matches(app, cr_app)), None)
            if not matching_cr_app:
                assert app._spec_uri is not None, 'Unexpected: app has no spec uri'
                self._unload_app(app._spec_uri)
                something_changed = True
        for cr_app_from_response in compute_resource_apps_from_response:
            matching_app = next((app for app in self._apps if _app_matches(app, cr_app_from_response)), None)
            if not matching_app:
                try:
                    self._load_app(cr_app_from_response)
                except Exception:
                    # print traceback
                    import traceback
                    traceback.print_exc()
                    print(f'Error loading app {cr_app_from_response.specUri}')
                    continue
                something_changed = True
        if something_changed:
            # Report the compute resource spec
            print('Reporting the compute resource spec')
            url_path = f'/api/compute_resource/compute_resources/{self._compute_resource_id}/spec'
            default_job_run_method = os.environ.get('DEFAULT_JOB_RUN_METHOD', 'local')
            available_job_run_methods_str = os.environ.get('AVAILABLE_JOB_RUN_METHODS', 'local')
            available_job_run_methods = [s.strip() for s in available_job_run_methods_str.split(',')]
            print(f'  default_job_run_method: {default_job_run_method}')
            print(f'  available_job_run_methods: {available_job_run_methods}')
            spec = {
                'apps': [a._spec_dict for a in self._apps],
                'defaultJobRunMethod': default_job_run_method,
                'availableJobRunMethods': available_job_run_methods
            }
            _compute_resource_put_api_request(
                url_path=url_path,
                compute_resource_id=self._compute_resource_id,
                compute_resource_private_key=self._compute_resource_private_key,
                data={
                    'spec': spec
                }
            )
    def find_app_with_processor(self, processor_name: str) -> Union[App, None]:
        for app in self._apps:
            for p in app._processors:
                if p._name == processor_name:
                    return app
        return None
    def _unload_app(self, spec_uri: str):
        print(f'Unloading app {spec_uri}')
        self._apps = [app for app in self._apps if app._spec_uri != spec_uri]
    def _load_app(self, cr_app: DendroComputeResourceApp):
        print(f'Loading app {cr_app.specUri}')
        app = App.from_spec_uri(
            spec_uri=cr_app.specUri,
        )
        self._apps.append(app)

        if 'aws_batch' in self._available_job_run_methods and app._app_image is not None:
            from ..aws_batch.aws_batch_job_definition import create_aws_batch_job_definition
            stack_id = 'DendroBatchStack'
            job_role_name = f"{stack_id}-BatchJobsAccessRole" # This must match with iac/aws_batch/aws_batch/stack_config.py
            # efs_fs_name = f"{stack_id}-EfsFileSystem" # This must match with iac/aws_batch/aws_batch/stack_config.py
            print(f'Creating AWS batch job definition for app {app._name}')
            environment_variables = []

            create_aws_batch_job_definition(
                dendro_app_name=app._name,
                dendro_app_image_uri=app._app_image,
                job_role_name=job_role_name,
                # efs_fs_name=efs_fs_name,
                environment_variables=environment_variables
            )

        processor_names_str = ', '.join([p._name for p in app._processors])
        print(f'  processors: {processor_names_str}')
        print('')
        print('')

def _app_matches(app: App, cr_app: DendroComputeResourceApp) -> bool:
    return (
        app._spec_uri == cr_app.specUri
    )



import os
import subprocess
from pathlib import Path


def test_app_processor_function(app_dir: str, processor: str, context: str):
    # Ensure the directory path is an absolute path
    app_dir_path = Path(app_dir).resolve()

    executable_path = str(app_dir_path / 'main.py')

    env = os.environ.copy()
    env['TEST_APP_PROCESSOR'] = '1'
    env['PROCESSOR_NAME'] = processor
    env['CONTEXT_FILE'] = context
    subprocess.run([executable_path], env=env)



from typing import List, Union
from pydantic import BaseModel


class FileManifestFile(BaseModel):
    name: str
    url: str
    size: Union[int, None] = None

class FileManifest(BaseModel):
    files: List[FileManifestFile]



import pytest
from dendro.common._api_request import _processor_get_api_request, _processor_put_api_request
from dendro.common._api_request import _client_get_api_request
from dendro.common._api_request import _gui_get_api_request, _gui_put_api_request, _gui_post_api_request, _gui_delete_api_request

@pytest.mark.api
def test_api_request_failures():
    from dendro.common._api_request import _use_api_test_client
    from dendro.mock import set_use_mock
    from test_integration import _get_fastapi_app
    from dendro.api_helpers.clients._get_mongo_client import _clear_mock_mongo_databases

    from fastapi.testclient import TestClient
    app = _get_fastapi_app()
    test_client = TestClient(app)
    _use_api_test_client(test_client)
    set_use_mock(True)

    try:
        # from requests import exceptions
        with pytest.raises(Exception):
            _processor_get_api_request(url_path='/api/incorrect', headers={})
        with pytest.raises(Exception):
            _processor_put_api_request(url_path='/api/incorrect', headers={}, data={})
        with pytest.raises(Exception):
            _client_get_api_request(url_path='/api/incorrect')
        with pytest.raises(Exception):
            _gui_get_api_request(url_path='/api/incorrect', github_access_token='incorrect')
        with pytest.raises(Exception):
            _gui_put_api_request(url_path='/api/incorrect', github_access_token='incorrect', data={})
        with pytest.raises(Exception):
            _gui_post_api_request(url_path='/api/incorrect', github_access_token='incorrect', data={})
        with pytest.raises(Exception):
            _gui_delete_api_request(url_path='/api/incorrect', github_access_token='incorrect')
    finally:
        _use_api_test_client(None)
        set_use_mock(False)
        _clear_mock_mongo_databases()



from typing import List, TYPE_CHECKING
from .SlurmJobHandler import SlurmJobHandler
from ..common.dendro_types import DendroJob
from ..sdk._run_job_parent_process import _set_job_status
if TYPE_CHECKING:
    from .AppManager import AppManager


# TODO: make these configurable
max_simultaneous_local_jobs = 2
max_simultaneous_aws_batch_jobs = 20

class JobManager:
    def __init__(self, *,
                 compute_resource_id: str,
                 compute_resource_private_key: str,
                 app_manager: 'AppManager'
                ):
        self._compute_resource_id = compute_resource_id
        self._compute_resource_private_key = compute_resource_private_key
        self._app_manager = app_manager

        # important to keep track of which jobs we attempted to start
        # so that we don't attempt multiple times in the case where starting failed
        self._attempted_to_start_job_ids = set()
        self._attempted_to_fail_job_ids = set()

        self._slurm_job_handler = SlurmJobHandler(job_manager=self)
    def handle_jobs(self, jobs: List[DendroJob]):
        local_jobs = [job for job in jobs if job.runMethod == 'local']
        aws_batch_jobs = [job for job in jobs if job.runMethod == 'aws_batch']
        slurm_jobs = [job for job in jobs if job.runMethod == 'slurm']
        none_jobs = [job for job in jobs if job.runMethod is None]

        for job in none_jobs:
            self._fail_job(job, 'runMethod is None')

        # Local jobs
        if 'local' in self._app_manager._available_job_run_methods:
            local_jobs_to_start = _choose_pending_jobs_to_start(local_jobs, max_simultaneous_local_jobs)
            for job in local_jobs_to_start:
                self._start_job(job)

        # AWS Batch jobs
        if 'aws_batch' in self._app_manager._available_job_run_methods:
            aws_jobs_to_start = _choose_pending_jobs_to_start(aws_batch_jobs, max_simultaneous_aws_batch_jobs)
            for job in aws_jobs_to_start:
                self._start_job(job)

        # SLURM jobs
        # this is more tricky... let's send the to slurm job handler
        if 'slurm' in self._app_manager._available_job_run_methods:
            self._slurm_job_handler.handle_jobs(slurm_jobs)
    def do_work(self):
        self._slurm_job_handler.do_work()
    def _job_is_pending(self, job: DendroJob) -> bool:
        return job.status == 'pending'
    def _start_job(self, job: DendroJob, run_process: bool = True, return_shell_command: bool = False):
        job_id = job.jobId
        if job_id in self._attempted_to_start_job_ids or job_id in self._attempted_to_fail_job_ids:
            return '' # see above comment about why this is necessary
        self._attempted_to_start_job_ids.add(job_id)
        job_private_key = job.jobPrivateKey
        processor_name = job.processorName
        app = self._app_manager.find_app_with_processor(processor_name)
        if app is None:
            msg = f'Could not find app with processor name {processor_name}'
            print(msg)
            _set_job_status(job_id=job_id, job_private_key=job_private_key, status='failed', error=msg)
            return ''
        try:
            print(f'Starting job {job_id} {processor_name}')
            from ._start_job import _start_job
            if job.requiredResources is None:
                raise Exception('Cannot start job... requiredResources is None')
            assert job.runMethod is not None, 'Unexpected: job.runMethod is None'
            return _start_job(
                job_id=job_id,
                job_private_key=job_private_key,
                processor_name=processor_name,
                run_method=job.runMethod,
                app=app,
                run_process=run_process,
                return_shell_command=return_shell_command,
                required_resources=job.requiredResources
            )
        except Exception as e: # pylint: disable=broad-except
            # do a traceback
            import traceback
            traceback.print_exc()
            msg = f'Failed to start job: {str(e)}'
            print(msg)
            self._fail_job(job, msg)
            return ''
    def _fail_job(self, job: DendroJob, error: str):
        job_id = job.jobId
        if job_id in self._attempted_to_fail_job_ids:
            return '' # see above comment about why this is necessary
        self._attempted_to_fail_job_ids.add(job_id)
        job_id = job.jobId
        job_private_key = job.jobPrivateKey
        print(f'Failing job {job_id}: {error}')
        _set_job_status(job_id=job_id, job_private_key=job_private_key, status='failed', error=error)

def _sort_jobs_by_timestamp_created(jobs: List[DendroJob]) -> List[DendroJob]:
    return sorted(jobs, key=lambda job: job.timestampCreated)

def _choose_pending_jobs_to_start(jobs: List[DendroJob], max_num_simultaneous_jobs: int) -> List[DendroJob]:
    num_non_pending_jobs = len([job for job in jobs if job.status != 'pending'])
    if num_non_pending_jobs < max_num_simultaneous_jobs:
        pending_jobs = [job for job in jobs if job.status == 'pending']
        pending_jobs = _sort_jobs_by_timestamp_created(pending_jobs)
        num_to_start = min(max_num_simultaneous_jobs - num_non_pending_jobs, len(pending_jobs))
        jobs_to_start = pending_jobs[:num_to_start]
    else:
        jobs_to_start = []
    return jobs_to_start



from typing import Optional
import os
import yaml
import time
from pathlib import Path
import shutil
import multiprocessing
from ..common._api_request import _compute_resource_get_api_request
from .register_compute_resource import env_var_keys
from ..common.dendro_types import DendroJob
from ..mock import using_mock
from .AppManager import AppManager
from .JobManager import JobManager


class Daemon:
    def __init__(self):
        self._compute_resource_id = os.getenv('COMPUTE_RESOURCE_ID', None)
        self._compute_resource_private_key = os.getenv('COMPUTE_RESOURCE_PRIVATE_KEY', None)
        if self._compute_resource_id is None:
            raise ValueError('Compute resource has not been initialized in this directory, and the environment variable COMPUTE_RESOURCE_ID is not set.')
        if self._compute_resource_private_key is None:
            raise ValueError('Compute resource has not been initialized in this directory, and the environment variable COMPUTE_RESOURCE_PRIVATE_KEY is not set.')

        available_job_run_methods_str = os.getenv('AVAILABLE_JOB_RUN_METHODS', 'local')
        available_job_run_methods = [s.strip() for s in available_job_run_methods_str.split(',') if s]
        for a in available_job_run_methods:
            if a not in ['local', 'aws_batch', 'slurm']:
                raise ValueError(f'Invalid job run method: {a}')

        self._app_manager = AppManager(
            compute_resource_id=self._compute_resource_id,
            compute_resource_private_key=self._compute_resource_private_key,
            available_job_run_methods=available_job_run_methods # type: ignore
        )
        self._app_manager.update_apps()

        self._job_manager = JobManager(
            compute_resource_id=self._compute_resource_id,
            compute_resource_private_key=self._compute_resource_private_key,
            app_manager=self._app_manager
        )

    def start(self, *, timeout: Optional[float] = None, cleanup_old_jobs=True): # timeout is used for testing
        timer_handle_jobs = 0

        time_scale_factor = 1 if not using_mock() else 10000

        assert self._compute_resource_id is not None
        assert self._compute_resource_private_key is not None

        print('Getting pubsub info')
        pubsub_subscription = get_pubsub_subscription(
            compute_resource_id=self._compute_resource_id,
            compute_resource_private_key=self._compute_resource_private_key
        )
        pubnub_subscribe_key = pubsub_subscription['pubnubSubscribeKey']
        if pubnub_subscribe_key != 'mock-subscribe-key':
            from .PubsubClient import PubsubClient
            pubsub_client = PubsubClient(
                pubnub_subscribe_key=pubnub_subscribe_key,
                pubnub_channel=pubsub_subscription['pubnubChannel'],
                pubnub_user=pubsub_subscription['pubnubUser'],
                compute_resource_id=self._compute_resource_id
            )
        else:
            pubsub_client = None

        # Start cleaning up old job directories
        # It's important to do this in a separate process
        # because it can take a long time to delete all the files in the tmp directories (remfile is the culprit)
        # and we don't want to block the main process from handling jobs
        if cleanup_old_jobs:
            cleanup_old_jobs_process = multiprocessing.Process(target=_cleanup_old_job_working_directories, args=(os.getcwd() + '/jobs',))
            cleanup_old_jobs_process.start()
        else:
            cleanup_old_jobs_process = None

        try:
            print('Starting compute resource')
            reported_that_compute_resource_is_running = False
            overall_timer = time.time()
            while True:
                elapsed_handle_jobs = time.time() - timer_handle_jobs
                # normally we will get pubsub messages for updates, but if we don't, we should check every 10 minutes
                time_to_handle_jobs = elapsed_handle_jobs > 60 * 10 / time_scale_factor
                messages = pubsub_client.take_messages() if pubsub_client is not None else []
                jobs_have_changed = False
                for msg in messages:
                    if msg['type'] == 'newPendingJob':
                        jobs_have_changed = True
                    if msg['type'] == 'jobStatusChanged':
                        jobs_have_changed = True
                    if msg['type'] == 'computeResourceAppsChanaged':
                        self._app_manager.update_apps()

                if time_to_handle_jobs or jobs_have_changed:
                    # if time_to_handle_jobs:
                    #     print('Checking for new jobs') # this pollutes the logs too much
                    timer_handle_jobs = time.time()
                    self._handle_jobs()

                self._job_manager.do_work()

                if not reported_that_compute_resource_is_running:
                    print('Compute resource is running')
                    reported_that_compute_resource_is_running = True

                overall_elapsed = time.time() - overall_timer
                if timeout is not None and overall_elapsed > timeout:
                    print(f'Compute resource timed out after {timeout} seconds')
                    return
                if overall_elapsed < 5 / time_scale_factor:
                    time.sleep(0.01 / time_scale_factor) # for the first few seconds we can sleep for a short time (useful for testing)
                else:
                    time.sleep(2 / time_scale_factor)
        finally:
            if cleanup_old_jobs_process is not None:
                cleanup_old_jobs_process.terminate()
            if pubsub_client is not None:
                pubsub_client.close() # unfortunately this doesn't actually stop the thread - it's a pubnub/python issue
    def _handle_jobs(self):
        url_path = f'/api/compute_resource/compute_resources/{self._compute_resource_id}/unfinished_jobs'
        if not self._compute_resource_id:
            return
        if not self._compute_resource_private_key:
            return
        try:
            resp = _compute_resource_get_api_request(
                url_path=url_path,
                compute_resource_id=self._compute_resource_id,
                compute_resource_private_key=self._compute_resource_private_key
            )
        except Exception as e:
            print(f'Error in compute resource get api request for {url_path}; {e}')
            return
        jobs = resp['jobs']
        jobs = [DendroJob(**job) for job in jobs]
        self._job_manager.handle_jobs(jobs)

def start_compute_resource(dir: str, *, timeout: Optional[float] = None, cleanup_old_jobs=True): # timeout is used for testing
    # Let's make sure pubnub is installed, because it's required for the daemon
    try:
        import pubnub # noqa
    except ImportError:
        raise ImportError('The pubnub package is not installed. You should use "pip install dendro[compute_resource]".')

    config_fname = os.path.join(dir, '.dendro-compute-resource-node.yaml')
    if os.path.exists(config_fname):
        with open(config_fname, 'r', encoding='utf8') as f:
            the_config = yaml.safe_load(f)
    else:
        the_config = {}
    for k in env_var_keys:
        if k in the_config and the_config[k]:
            os.environ[k] = the_config[k]
        else:
            the_config[k] = ''
    for k in the_config.keys():
        if k not in env_var_keys:
            print('WARNING: Unknown key in config file: ' + k)
    with open(config_fname, 'w', encoding='utf8') as f:
        yaml.dump(the_config, f)
    daemon = Daemon()
    daemon.start(timeout=timeout, cleanup_old_jobs=cleanup_old_jobs)

def get_pubsub_subscription(*, compute_resource_id: str, compute_resource_private_key: str):
    url_path = f'/api/compute_resource/compute_resources/{compute_resource_id}/pubsub_subscription'
    resp = _compute_resource_get_api_request(
        url_path=url_path,
        compute_resource_id=compute_resource_id,
        compute_resource_private_key=compute_resource_private_key
    )
    return resp['subscription']

def _cleanup_old_job_working_directories(dir: str):
    """Delete working dirs that are more than 24 hours old"""
    jobs_dir = Path(dir)
    while True:
        if not jobs_dir.exists():
            continue
        for job_dir in jobs_dir.iterdir():
            if job_dir.is_dir():
                elapsed = time.time() - job_dir.stat().st_mtime
                if elapsed > 24 * 60 * 60:
                    print(f'Removing old working dir {job_dir}')
                    shutil.rmtree(job_dir)
        time.sleep(60)



from typing import List, Union, Type, get_type_hints
import os
import json
from .AppProcessor import AppProcessor
from ._run_job_parent_process import _run_job_parent_process
from ._run_job_child_process import _run_job_child_process
from ._load_spec_from_uri import _load_spec_from_uri
from .ProcessorBase import ProcessorBase
from .InputFile import InputFile
from .InputFolder import InputFolder
from .OutputFile import OutputFile
from .OutputFolder import OutputFolder


class DendroAppException(Exception):
    pass

class App:
    """An app"""
    def __init__(
        self,
        name: str,
        *,
        description: str,
        app_image: Union[str, None] = None,
        app_executable: Union[str, None] = None
    ) -> None:
        """Construct a new Dendro App

        Args:
            name (str): The name of the app
            description (str): A description of the app
            app_image (Union[str, None], optional): The URI for the docker image, or None if no image is to be used. Defaults to None.
            app_executable (Union[str, None], optional): The app executable within the docker image, or on the local file system if app_image=None. If app_image is set, this will default to /app/main.py. If app_image is not set, then an exception with be thrown if this is not set.
        """
        if not app_image:
            if not app_executable:
                raise DendroAppException('You must set app_executable if app_image is not set')

        if len(description) > 1000:
            raise Exception('Description for app must be less than 1000 characters')

        self._name = name
        self._description = description
        self._app_image = app_image
        self._app_executable = app_executable
        self._processors: List[AppProcessor] = []

        self._spec_dict: Union[dict, None] = None # this is set when the app is loaded from a spec (internal use only)
        self._spec_uri: Union[str, None] = None # this is set when the app is loaded from a spec_uri (internal use only)

    def add_processor(self, processor_class: Type[ProcessorBase]):
        """Add a processor to the app

        Args:
            processor_class (Type[ProcessorBase]): The processor class for the processor
        """
        P = AppProcessor.from_processor_class(processor_class)
        self._processors.append(P)

    def run(self):
        """This function should be called once in main.py"""
        SPEC_OUTPUT_FILE = os.environ.get('SPEC_OUTPUT_FILE', None)
        if SPEC_OUTPUT_FILE is not None:
            if os.environ.get('JOB_ID', None) is not None:
                raise Exception('Cannot set both JOB_ID and SPEC_OUTPUT_FILE')
            with open(SPEC_OUTPUT_FILE, 'w') as f:
                json.dump(self.get_spec(), f, indent=4)
            return
        JOB_ID = os.environ.get('JOB_ID', None)
        if JOB_ID is not None:
            JOB_PRIVATE_KEY = os.environ.get('JOB_PRIVATE_KEY', None)
            JOB_INTERNAL = os.environ.get('JOB_INTERNAL', None)
            APP_EXECUTABLE = os.environ.get('APP_EXECUTABLE', None)
            JOB_TIMEOUT_SEC = os.environ.get('JOB_TIMEOUT_SEC', None)
            if JOB_TIMEOUT_SEC is not None:
                JOB_TIMEOUT_SEC = int(JOB_TIMEOUT_SEC)
            if JOB_PRIVATE_KEY is None:
                raise KeyError('JOB_PRIVATE_KEY is not set')
            if JOB_INTERNAL == '1':
                # In this mode, we run the job directly
                # This is called internally by the other run mode (need to explain this better)
                return _run_job_child_process(job_id=JOB_ID, job_private_key=JOB_PRIVATE_KEY, processors=self._processors)

            # In this mode we run the job, including the top-level interactions with the dendro API, such as setting the status and the console output, and checking whether the job has been canceled/deleted
            if APP_EXECUTABLE is None:
                raise KeyError('APP_EXECUTABLE is not set')
            return _run_job_parent_process(
                job_id=JOB_ID,
                job_private_key=JOB_PRIVATE_KEY,
                app_executable=APP_EXECUTABLE,
                job_timeout_sec=JOB_TIMEOUT_SEC
            )
        TEST_APP_PROCESSOR = os.environ.get('TEST_APP_PROCESSOR', None)
        if TEST_APP_PROCESSOR is not None:
            PROCESSOR_NAME = os.environ.get('PROCESSOR_NAME', None)
            CONTEXT_FILE = os.environ.get('CONTEXT_FILE', None)
            if PROCESSOR_NAME is None:
                raise KeyError('PROCESSOR_NAME is not set')
            if CONTEXT_FILE is None:
                raise KeyError('CONTEXT_FILE is not set')
            if CONTEXT_FILE.endswith('.json'):
                with open(CONTEXT_FILE, 'r') as f:
                    context = json.load(f)
            elif CONTEXT_FILE.endswith('.yml') or CONTEXT_FILE.endswith('.yaml'):
                import yaml
                with open(CONTEXT_FILE, 'r') as f:
                    context = yaml.safe_load(f)
            else:
                raise Exception(f'Unrecognized file extension: {CONTEXT_FILE}')
            processor = next((p for p in self._processors if p._name == PROCESSOR_NAME), None)
            if not processor:
                raise Exception(f'Processor not found: {PROCESSOR_NAME}')

            # handle input files coming in as url's
            for input in processor._inputs:
                input_value = _get_in_dict_where_key_may_have_dots(context, input.name)
                if input_value.startswith('http://') or input_value.startswith('https://'):
                    _set_in_dict_where_key_may_have_dots(context, input.name, InputFile(name=input.name, url=input_value))
                elif input_value.startswith('/') or input_value.startswith('./') or input_value.startswith('../'):
                    _set_in_dict_where_key_may_have_dots(context, input.name, InputFile(name=input.name, local_file_name=input_value))
                else:
                    raise Exception(f'Input value for {input.name} must either be a URL or a local path starting with /, ./ or ../. Got: {input_value}')

            # handle input folders coming in as url's
            for input_folder in processor._input_folders:
                input_folder_value = _get_in_dict_where_key_may_have_dots(context, input_folder.name)
                if input_folder_value.startswith('http://') or input_folder_value.startswith('https://'):
                    _set_in_dict_where_key_may_have_dots(context, input_folder.name, InputFolder(name=input_folder.name, url=input_folder_value))
                elif input_folder_value.startswith('/') or input_folder_value.startswith('./') or input_folder_value.startswith('../'):
                    _set_in_dict_where_key_may_have_dots(context, input_folder.name, InputFolder(name=input_folder.name, local_folder_name=input_folder_value))
                else:
                    raise Exception(f'Input folder value for {input_folder.name} must either be a URL or a local path starting with /, ./ or ../. Got: {input_folder_value}')

            # handle output files coming in as file paths
            for output in processor._outputs:
                output_value = _get_in_dict_where_key_may_have_dots(context, output.name)
                if output_value.startswith('https://') or output_value.startswith('http://'):
                    raise Exception(f'Output value for {output.name} cannot be a URL. Should be local file path. Got: {output_value}')
                elif output_value.startswith('/') or output_value.startswith('./') or output_value.startswith('../'):
                    _set_in_dict_where_key_may_have_dots(context, output.name, OutputFile(name=output.name, output_file_name=output_value))
                else:
                    raise Exception(f'Output value for {output.name} must be a local path starting with /, ./ or ../. Got: {output_value}')

            # handle output folders coming in as file paths
            for output_folder in processor._output_folders:
                output_folder_value = _get_in_dict_where_key_may_have_dots(context, output_folder.name)
                if output_folder_value.startswith('https://') or output_folder_value.startswith('http://'):
                    raise Exception(f'Output folder value for {output_folder.name} cannot be a URL. Should be local folder path. Got: {output_folder_value}')
                elif output_folder_value.startswith('/') or output_folder_value.startswith('./') or output_folder_value.startswith('../'):
                    _set_in_dict_where_key_may_have_dots(context, output_folder.name, OutputFolder(name=output_folder.name, output_folder_name=output_folder_value))
                else:
                    raise Exception(f'Output folder value for {output_folder.name} must be a local path starting with /, ./ or ../. Got: {output_folder_value}')

            processor_class = processor._processor_class
            assert processor_class, f'Processor does not have a processor_class: {PROCESSOR_NAME}'
            context_type = _get_type_of_context_in_processor_class(processor_class)
            assert context_type, f'Processor does not have a context type: {PROCESSOR_NAME}'
            context = context_type(**context)
            processor_class.run(context)
            return
        raise KeyError('You must set JOB_ID as an environment variable to run a job')

    def make_spec_file(self, spec_output_file: str = 'spec.json'):
        """Create a spec.json file. This is called internally."""
        with open(spec_output_file, 'w', encoding='utf-8') as f:
            json.dump(self.get_spec(), f, indent=4)

    def get_spec(self):
        """Get the spec for this app. This is called internally."""
        processors = []
        for processor in self._processors:
            processors.append(
                processor.get_spec()
            )
        spec = {
            'name': self._name,
            'description': self._description,
            'appImage': self._app_image,
            'appExecutable': self._app_executable,
            'executable': self._app_executable,
            'processors': processors
        }
        return spec

    @staticmethod
    def from_spec(spec):
        """Define an app from a spec. This is called internally."""
        app = App(
            name=spec['name'],
            description=spec['description'],
            app_image=spec.get('appImage', None),
            app_executable=spec.get('appExecutable', None),
        )
        app._spec_dict = spec
        for processor_spec in spec['processors']:
            processor = AppProcessor.from_spec(processor_spec)
            app._processors.append(processor)
        return app

    @staticmethod
    def from_spec_uri(
        spec_uri: str
    ):
        """Define an app from a spec URI (e.g., a gh url to the spec.json blob). This is called internally."""
        spec: dict = _load_spec_from_uri(spec_uri)
        a = App.from_spec(spec)
        a._spec_uri = spec_uri
        return a

def _get_type_of_context_in_processor_class(processor_class):
    # Retrieve the 'run' method from the processor class
    run_method = getattr(processor_class, 'run', None)
    if not run_method:
        raise Exception(f'Processor class does not have a run method: {processor_class}')
    # Get the type hints of the 'run' method
    type_hints = get_type_hints(run_method)
    # Return the type hint for the 'context' parameter
    return type_hints.get('context')

def _get_in_dict_where_key_may_have_dots(d: dict, key: str):
    if '.' in key:
        key_parts = key.split('.')
        for key_part in key_parts[:-1]:
            d = d[key_part]
        return d[key_parts[-1]]
    else:
        return d[key]

def _set_in_dict_where_key_may_have_dots(d: dict, key: str, value):
    if '.' in key:
        key_parts = key.split('.')
        for key_part in key_parts[:-1]:
            d = d[key_part]
        d[key_parts[-1]] = value
    else:
        d[key] = value



def sign_message(msg: dict, public_key_hex: str, private_key_hex: str) -> str:
    return _sign_message(msg, public_key_hex, private_key_hex)

ed25519PubKeyPrefix = "302a300506032b6570032100"
ed25519PrivateKeyPrefix = "302e020100300506032b657004220420"

def _deterministic_json_dumps(x: dict):
    import simplejson
    return simplejson.dumps(x, separators=(',', ':'), indent=None, allow_nan=False, sort_keys=True)

def _sha1_of_string(txt: str) -> str:
    import hashlib
    hh = hashlib.sha1(txt.encode('utf-8'))
    ret = hh.hexdigest()
    return ret

def _sign_message(msg: dict, public_key_hex: str, private_key_hex: str) -> str:
    msg_json = _deterministic_json_dumps(msg)
    return _sign_message_str(msg_json, public_key_hex, private_key_hex)


def _sign_message_str(msg: str, public_key_hex: str, private_key_hex: str) -> str:
    from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey
    msg_hash = _sha1_of_string(msg)
    msg_bytes = bytes.fromhex(msg_hash)
    privk = Ed25519PrivateKey.from_private_bytes(bytes.fromhex(private_key_hex))
    pubk = Ed25519PublicKey.from_public_bytes(bytes.fromhex(public_key_hex))
    signature = privk.sign(msg_bytes).hex()
    pubk.verify(bytes.fromhex(signature), msg_bytes)
    return signature

def _verify_signature(msg: dict, public_key_hex: str, signature: str):
    msg_json = _deterministic_json_dumps(msg)
    return _verify_signature_str(msg_json, public_key_hex, signature)

def _verify_signature_str(msg: str, public_key_hex: str, signature: str):
    from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
    msg_hash = _sha1_of_string(msg)
    msg_bytes = bytes.fromhex(msg_hash)
    pubk = Ed25519PublicKey.from_public_bytes(bytes.fromhex(public_key_hex))
    try:
        pubk.verify(bytes.fromhex(signature), msg_bytes)
    except: # noqa: E722
        return False
    return True

def generate_keypair():
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
    privk = Ed25519PrivateKey.generate()
    pubk = privk.public_key()
    private_key_hex = privk.private_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PrivateFormat.Raw,
        encryption_algorithm=serialization.NoEncryption()
    ).hex() # type: ignore
    public_key_hex = pubk.public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw
    ).hex() # type: ignore
    test_msg = {'a': 1}
    test_signature = _sign_message(test_msg, public_key_hex, private_key_hex)
    assert _verify_signature(test_msg, public_key_hex, test_signature)
    return public_key_hex, private_key_hex



from typing import List, Union
from .... import BaseModel
from fastapi import APIRouter, Header
from ....common.dendro_types import DendroProject, DendroFile, DendroJob
from ...clients.db import fetch_project, fetch_project_files, fetch_project_jobs, fetch_compute_resource
from ..common import api_route_wrapper
from ....common.dendro_types import CreateJobRequest, CreateJobResponse, DendroComputeResource
from ..gui.create_job_route import create_job_handler
from ...core.settings import get_settings

router = APIRouter()

# get project
class GetProjectResponse(BaseModel):
    project: DendroProject
    success: bool

class ProjectError(Exception):
    pass

@router.get("/projects/{project_id}")
@api_route_wrapper
async def get_project(project_id) -> GetProjectResponse:
    project = await fetch_project(project_id)
    if project is None:
        raise ProjectError(f"No project with ID {project_id}")
    if not project.computeResourceId:
        project.computeResourceId = get_settings().DEFAULT_COMPUTE_RESOURCE_ID
    return GetProjectResponse(project=project, success=True)

# get project files
class GetProjectFilesResponse(BaseModel):
    files: List[DendroFile]
    success: bool

@router.get("/projects/{project_id}/files")
@api_route_wrapper
async def get_project_files(project_id) -> GetProjectFilesResponse:
    files = await fetch_project_files(project_id)
    return GetProjectFilesResponse(files=files, success=True)

# get project jobs
class GetProjectJobsResponse(BaseModel):
    jobs: List[DendroJob]
    success: bool

@router.get("/projects/{project_id}/jobs")
@api_route_wrapper
async def get_project_jobs(project_id) -> GetProjectJobsResponse:
    jobs = await fetch_project_jobs(project_id)
    return GetProjectJobsResponse(jobs=jobs, success=True)

@router.post("/jobs")
@api_route_wrapper
async def create_job(
    data: CreateJobRequest,
    dendro_api_key: Union[str, None] = Header(None)
) -> CreateJobResponse:
    return await create_job_handler(
        data=data,
        dendro_api_key=dendro_api_key,
        github_access_token=None
    )

# get compute resource
class GetComputeResourceResponse(BaseModel):
    computeResource: DendroComputeResource
    success: bool

class ComputeResourceNotFoundException(Exception):
    pass

@router.get("/compute_resources/{compute_resource_id}")
@api_route_wrapper
async def get_compute_resource(compute_resource_id) -> GetComputeResourceResponse:
    compute_resource = await fetch_compute_resource(compute_resource_id, raise_on_not_found=True)
    assert compute_resource
    return GetComputeResourceResponse(computeResource=compute_resource, success=True)



import os
import pytest
from dendro.api_helpers.core._get_project_role import _check_user_can_edit_project, _check_user_can_read_project, _check_user_is_project_admin, _project_has_user
from dendro.common.dendro_types import DendroProject, DendroProjectUser

def test_get_project_role():
    old_env = os.environ.copy()
    try:
        os.environ['ADMIN_USER_IDS'] = '["github|the-system-admin"]'
        project = DendroProject(
            projectId='test-project-id',
            name='test-project-name',
            description='test-project-description',
            ownerId='github|test-owner-id',
            users=[
                DendroProjectUser(
                    userId='github|admin-user',
                    role='admin'
                ),
                DendroProjectUser(
                    userId='github|editor-user',
                    role='editor'
                ),
                DendroProjectUser(
                    userId='github|viewer-user',
                    role='viewer'
                )
            ],
            publiclyReadable=False,
            tags=[],
            timestampCreated=0,
            timestampModified=0,
            computeResourceId=None
        )
        _check_user_can_read_project(project, 'github|admin-user')
        _check_user_can_read_project(project, 'github|editor-user')
        _check_user_can_read_project(project, 'github|viewer-user')
        _check_user_can_read_project(project, 'github|the-system-admin')

        with pytest.raises(Exception):
            _check_user_can_read_project(project, 'github|non-user')
        with pytest.raises(Exception):
            _check_user_can_read_project(project, None)
        _check_user_can_edit_project(project, 'github|admin-user')
        _check_user_can_edit_project(project, 'github|editor-user')
        _check_user_can_edit_project(project, 'github|the-system-admin')

        with pytest.raises(Exception):
            _check_user_can_edit_project(project, 'github|viewer-user')
        with pytest.raises(Exception):
            _check_user_can_edit_project(project, 'github|non-user')
        with pytest.raises(Exception):
            _check_user_can_edit_project(project, None)
        _check_user_is_project_admin(project, 'github|admin-user')
        with pytest.raises(Exception):
            _check_user_is_project_admin(project, 'github|editor-user')
        with pytest.raises(Exception):
            _check_user_is_project_admin(project, 'github|viewer-user')
        with pytest.raises(Exception):
            _check_user_is_project_admin(project, 'github|non-user')
        with pytest.raises(Exception):
            _check_user_is_project_admin(project, None)
        _check_user_is_project_admin(project, 'github|the-system-admin')

        assert _project_has_user(project, 'github|admin-user')
        assert _project_has_user(project, 'github|editor-user')
        assert _project_has_user(project, 'github|viewer-user')
        assert not _project_has_user(project, 'github|non-user')
        assert not _project_has_user(project, None)
        assert not _project_has_user(project, 'github|the-system-admin')

        project.publiclyReadable = True
        _check_user_can_read_project(project, 'github|admin-user')
        _check_user_can_read_project(project, 'github|editor-user')
        _check_user_can_read_project(project, 'github|viewer-user')
        _check_user_can_read_project(project, 'github|non-user')
        _check_user_can_read_project(project, None)
        _check_user_can_read_project(project, 'github|the-system-admin')

        _check_user_can_edit_project(project, 'github|admin-user')
        _check_user_can_edit_project(project, 'github|editor-user')
        with pytest.raises(Exception):
            _check_user_can_edit_project(project, 'github|viewer-user')
        with pytest.raises(Exception):
            _check_user_can_edit_project(project, 'github|non-user')
        with pytest.raises(Exception):
            _check_user_can_edit_project(project, None)
    finally:
        os.environ = old_env



from typing import List
from .... import BaseModel
from fastapi import APIRouter, Header
from ...services._crypto_keys import _verify_signature_str
from ....common.dendro_types import DendroComputeResourceApp, DendroFile, DendroJob, ComputeResourceSpec, PubsubSubscription
from ...clients.db import fetch_compute_resource, fetch_compute_resource_jobs, fetch_multi_project_files, set_compute_resource_spec
from ...core.settings import get_settings
from ....mock import using_mock
from ..common import api_route_wrapper

router = APIRouter()

# get apps
class GetAppsResponse(BaseModel):
    apps: List[DendroComputeResourceApp]
    success: bool

class ComputeResourceNotFoundException(Exception):
    pass

@router.get("/compute_resources/{compute_resource_id}/apps")
@api_route_wrapper
async def compute_resource_get_apps(
    compute_resource_id: str,
    compute_resource_payload: str = Header(...),
    compute_resource_signature: str = Header(...)
) -> GetAppsResponse:
    # authenticate the request
    expected_payload = f'/api/compute_resource/compute_resources/{compute_resource_id}/apps'
    _authenticate_compute_resource_request(
        compute_resource_id=compute_resource_id,
        compute_resource_payload=compute_resource_payload,
        compute_resource_signature=compute_resource_signature,
        expected_payload=expected_payload
    )

    compute_resource = await fetch_compute_resource(compute_resource_id, raise_on_not_found=True)
    assert compute_resource
    apps = compute_resource.apps
    return GetAppsResponse(apps=apps, success=True)

# get pubsub subscription
class GetPubsubSubscriptionResponse(BaseModel):
    subscription: PubsubSubscription
    success: bool

@router.get("/compute_resources/{compute_resource_id}/pubsub_subscription")
@api_route_wrapper
async def compute_resource_get_pubsub_subscription(
    compute_resource_id: str,
    compute_resource_payload: str = Header(...),
    compute_resource_signature: str = Header(...)
):
    # authenticate the request
    expected_payload = f'/api/compute_resource/compute_resources/{compute_resource_id}/pubsub_subscription'
    _authenticate_compute_resource_request(
        compute_resource_id=compute_resource_id,
        compute_resource_payload=compute_resource_payload,
        compute_resource_signature=compute_resource_signature,
        expected_payload=expected_payload
    )

    compute_resource = await fetch_compute_resource(compute_resource_id, raise_on_not_found=True)
    assert compute_resource
    if using_mock():
        subscription = PubsubSubscription(
            pubnubSubscribeKey='mock-subscribe-key',
            pubnubChannel=compute_resource_id,
            pubnubUser=compute_resource_id
        )
        return GetPubsubSubscriptionResponse(subscription=subscription, success=True)
    else: # pragma: no cover
        VITE_PUBNUB_SUBSCRIBE_KEY = get_settings().PUBNUB_SUBSCRIBE_KEY
        if VITE_PUBNUB_SUBSCRIBE_KEY is None:
            raise KeyError('Environment variable not set: VITE_PUBNUB_SUBSCRIBE_KEY (compute_resource)')
        subscription = PubsubSubscription(
            pubnubSubscribeKey=VITE_PUBNUB_SUBSCRIBE_KEY,
            pubnubChannel=compute_resource_id,
            pubnubUser=compute_resource_id
        )
        return GetPubsubSubscriptionResponse(subscription=subscription, success=True)

# get unfinished jobs
class GetUnfinishedJobsResponse(BaseModel):
    jobs: List[DendroJob]
    success: bool

@router.get("/compute_resources/{compute_resource_id}/unfinished_jobs")
@api_route_wrapper
async def compute_resource_get_unfinished_jobs(
    compute_resource_id: str,
    compute_resource_payload: str = Header(...),
    compute_resource_signature: str = Header(...),
) -> GetUnfinishedJobsResponse:
    # authenticate the request
    expected_payload = f'/api/compute_resource/compute_resources/{compute_resource_id}/unfinished_jobs'
    _authenticate_compute_resource_request(
        compute_resource_id=compute_resource_id,
        compute_resource_payload=compute_resource_payload,
        compute_resource_signature=compute_resource_signature,
        expected_payload=expected_payload
    )

    jobs = await fetch_compute_resource_jobs(compute_resource_id, statuses=['pending', 'queued', 'starting', 'running'], exclude_those_pending_approval=True, include_private_keys=True)

    # exclude those for which the input files are pending
    relevant_project_ids: list[str] = []
    for job in jobs:
        if job.projectId not in relevant_project_ids:
            relevant_project_ids.append(job.projectId)
    pending_output_files: List[DendroFile] = await fetch_multi_project_files(project_ids=relevant_project_ids, pending_only=True)
    pending_output_file_names = set([f.fileName for f in pending_output_files])
    ready_jobs: List[DendroJob] = []
    for job in jobs:
        okay = True
        for input_file in job.inputFiles:
            if input_file.fileName in pending_output_file_names:
                okay = False
                break
        if okay:
            ready_jobs.append(job)

    return GetUnfinishedJobsResponse(jobs=ready_jobs, success=True)

# set spec
class SetSpecRequest(BaseModel):
    spec: ComputeResourceSpec

class SetSpecResponse(BaseModel):
    success: bool

@router.put("/compute_resources/{compute_resource_id}/spec")
@api_route_wrapper
async def compute_resource_set_spec(
    compute_resource_id,
    data: SetSpecRequest,
    compute_resource_payload: str = Header(...),
    compute_resource_signature: str = Header(...)
) -> SetSpecResponse:
    # authenticate the request
    expected_payload = f'/api/compute_resource/compute_resources/{compute_resource_id}/spec'
    _authenticate_compute_resource_request(
        compute_resource_id=compute_resource_id,
        compute_resource_payload=compute_resource_payload,
        compute_resource_signature=compute_resource_signature,
        expected_payload=expected_payload
    )

    spec = data.spec

    await set_compute_resource_spec(compute_resource_id, spec)

    return SetSpecResponse(success=True)

class UnexpectedException(Exception):
    pass

class InvalidSignatureException(Exception):
    pass


def _authenticate_compute_resource_request(
    compute_resource_id: str,
    compute_resource_payload: str,
    compute_resource_signature: str,
    expected_payload: str
):
    if compute_resource_payload != expected_payload:
        raise UnexpectedException('Unexpected payload')
    if not _verify_signature_str(compute_resource_payload, compute_resource_id, compute_resource_signature):
        raise InvalidSignatureException('Invalid signature')



from typing import Union, List, Any, Optional, Literal

from .. import BaseModel

class DendroProjectUser(BaseModel):
    userId: str
    role: str # 'admin' | 'editor' | 'viewer'

class DendroProject(BaseModel):
    projectId: str
    name: str
    description: str
    ownerId: str
    users: List[DendroProjectUser]
    publiclyReadable: bool
    tags: List[str]
    timestampCreated: float
    timestampModified: float
    computeResourceId: Union[str, None] = None # it seems this needs to go at the end, otherwise it will be required by pydantic - not sure why
    analysisSourceUrl: Union[str, None] = None

class DendroJobInputFile(BaseModel):
    name: str
    fileId: str
    fileName: str
    isFolder: Union[bool, None] = None

class DendroJobInputParameter(BaseModel):
    name: str
    value: Union[Any, None] = None
    secret: Union[bool, None] = None

class DendroJobOutputFile(BaseModel):
    name: str
    fileName: str
    fileId: Union[str, None] = None
    isFolder: Union[bool, None] = None

class ComputeResourceSpecProcessorParameter(BaseModel):
    name: str
    description: str
    type: str
    default: Union[Any, None] = None
    options: Union[List[str], List[int], None] = None
    secret: bool = False

class ComputeResourceSpecProcessorInput(BaseModel):
    name: str
    description: str
    list: bool = False

class ComputeResourceSpecProcessorInputFolder(BaseModel):
    name: str
    description: str
    list: bool = False

class ComputeResourceSpecProcessorOutput(BaseModel):
    name: str
    description: str

class ComputeResourceSpecProcessorOutputFolder(BaseModel):
    name: str
    description: str

class ComputeResourceSpecProcessorAttribute(BaseModel):
    name: str
    value: Any

class ComputeResourceSpecProcessorTag(BaseModel):
    tag: str

class ComputeResourceSpecProcessor(BaseModel):
    name: str
    description: str
    inputs: List[ComputeResourceSpecProcessorInput]
    inputFolders: Union[List[ComputeResourceSpecProcessorInputFolder], None] = None
    outputs: List[ComputeResourceSpecProcessorOutput]
    outputFolders: Union[List[ComputeResourceSpecProcessorOutputFolder], None] = None
    parameters: List[ComputeResourceSpecProcessorParameter]
    attributes: List[ComputeResourceSpecProcessorAttribute]
    tags: List[ComputeResourceSpecProcessorTag]

class DendroJobRequiredResources(BaseModel):
    numCpus: int
    numGpus: int
    memoryGb: float
    timeSec: float

class DendroJobUsedResources(BaseModel):
    numCpus: int
    numGpus: int
    memoryGb: float
    timeSec: float

class DendroJob(BaseModel):
    projectId: str
    jobId: str
    jobPrivateKey: str
    userId: str
    processorName: str
    inputFiles: List[DendroJobInputFile]
    inputFileIds: List[str]
    inputParameters: List[DendroJobInputParameter]
    outputFiles: List[DendroJobOutputFile]
    requiredResources: Union[DendroJobRequiredResources, None] = None
    usedResources: Union[DendroJobUsedResources, None] = None
    runMethod: Union[Literal['local', 'aws_batch', 'slurm'], None] = None
    timestampCreated: float
    computeResourceId: str
    status: str # 'pending' | 'queued' | 'starting' | 'running' | 'completed' | 'failed'
    batchId: Union[str, None] = None
    error: Union[str, None] = None
    processorVersion: Union[str, None] = None
    computeResourceNodeId: Union[str, None] = None # obsolete
    computeResourceNodeName: Union[str, None] = None # obsolete
    consoleOutputUrl: Union[str, None] = None
    resourceUtilizationLogUrl: Union[str, None] = None
    timestampQueued: Union[float, None] = None
    timestampStarting: Union[float, None] = None
    timestampStarted: Union[float, None] = None
    timestampFinished: Union[float, None] = None
    outputFileIds: Union[List[str], None] = None
    processorSpec: ComputeResourceSpecProcessor
    dandiApiKey: Union[str, None] = None
    deleted: Union[bool, None] = None
    pendingApproval: Union[bool, None] = None

class DendroFile(BaseModel):
    projectId: str
    fileId: str
    userId: str
    fileName: str
    size: int
    timestampCreated: float
    content: str # e.g., 'url:https://...'
    metadata: dict
    isFolder: Union[bool, None] = None
    jobId: Union[str, None] = None # the job that produced this file

# obsolete
class ComputeResourceAwsBatchOpts(BaseModel):
    jobQueue: Optional[str] = None # obsolete
    jobDefinition: Optional[str] = None # obsolete
    useAwsBatch: Optional[bool] = None # obsolete

# obsolete
class ComputeResourceSlurmOpts(BaseModel):
    partition: Union[str, None] = None
    time: Union[str, None] = None
    cpusPerTask: Union[int, None] = None
    otherOpts: Union[str, None] = None

class DendroComputeResourceApp(BaseModel):
    name: str
    specUri: str
    executablePath: Union[str, None] = None # to be removed (once database has been cleared)
    container: Union[str, None] = None # to be removed (once database has been cleared)
    awsBatch: Union[ComputeResourceAwsBatchOpts, None] = None # obsolete
    slurm: Union[ComputeResourceSlurmOpts, None] = None # obsolete

class ComputeResourceSpecApp(BaseModel):
    name: str
    description: str
    processors: List[ComputeResourceSpecProcessor]
    appImage: Union[str, None] = None
    appExecutable: Union[str, None] = None

class ComputeResourceSpec(BaseModel):
    apps: List[ComputeResourceSpecApp]
    defaultJobRunMethod: Union[Literal['local', 'aws_batch', 'slurm'], None] = None
    availableJobRunMethods: Union[List[Literal['local', 'aws_batch', 'slurm']], None] = None

class DendroComputeResource(BaseModel):
    computeResourceId: str
    ownerId: str
    name: str
    timestampCreated: float
    apps: List[DendroComputeResourceApp]
    spec: Union[ComputeResourceSpec, None] = None

class PubsubSubscription(BaseModel):
    pubnubSubscribeKey: str
    pubnubChannel: str
    pubnubUser: str

class ProcessorGetJobResponseInput(BaseModel):
    name: str
    url: str

class ProcessorGetJobResponseInputFolderFile(BaseModel):
    name: str
    url: str
    size: Union[int, None] = None

class ProcessorGetJobResponseInputFolder(BaseModel):
    name: str
    files: List[ProcessorGetJobResponseInputFolderFile]

class ProcessorGetJobResponseOutput(BaseModel):
    name: str

class ProcessorGetJobResponseOutputFolder(BaseModel):
    name: str

class ProcessorGetJobResponseParameter(BaseModel):
    name: str
    value: Any

class ProcessorGetJobResponse(BaseModel):
    jobId: str
    status: str
    processorName: str
    inputs: List[ProcessorGetJobResponseInput]
    inputFolders: Union[List[ProcessorGetJobResponseInputFolder], None] = None
    outputs: List[ProcessorGetJobResponseOutput]
    outputFolders: Union[List[ProcessorGetJobResponseOutputFolder], None] = None
    parameters: List[ProcessorGetJobResponseParameter]

class DendroUser(BaseModel):
    userId: str
    dendroApiKey: Union[str, None] = None

class CreateJobRequestInputFile(BaseModel):
    name: str
    fileName: str
    isFolder: Union[bool, None] = None

class CreateJobRequestOutputFile(BaseModel):
    name: str
    fileName: str
    isFolder: Union[bool, None] = None

class CreateJobRequestInputParameter(BaseModel):
    name: str
    value: Union[Any, None]

class CreateJobRequest(BaseModel):
    projectId: str
    processorName: str
    inputFiles: List[CreateJobRequestInputFile]
    outputFiles: List[CreateJobRequestOutputFile]
    inputParameters: List[CreateJobRequestInputParameter]
    processorSpec: ComputeResourceSpecProcessor
    batchId: Union[str, None] = None
    dandiApiKey: Union[str, None] = None
    requiredResources: DendroJobRequiredResources
    runMethod: Literal['local', 'aws_batch', 'slurm']

class CreateJobResponse(BaseModel):
    jobId: str
    success: bool

class ComputeResourceUserUsage(BaseModel):
    computeResourceId: str
    userId: str
    jobsIncludingDeleted: List[DendroJob]



from typing import Union
import shutil
import requests
from ..mock import using_mock
from pydantic import BaseModel


class SetOutputFileException(Exception):
    pass

class OutputFile(BaseModel):
    name: Union[str, None] = None # the name of the output within the context of the processor (not needed when output_file_name is specified for local testing)
    job_id: Union[str, None] = None
    job_private_key: Union[str, None] = None
    was_uploaded: bool = False
    output_file_name: Union[str, None] = None # for local testing
    def set(self, local_file_name: str):
        print('output.set() is deprecated. Please use output.upload() instead.')
        self.upload(local_file_name)
    def upload(self, local_file_name: str):
        from .Job import _get_upload_url_for_output_file # avoid circular import

        if self.output_file_name is not None:
            if self.job_id is not None:
                raise Exception('Cannot specify both output_file_name and job_id in OutputFile')
            if self.job_private_key is not None:
                raise Exception('Cannot specify both output_file_name and job_private_key in OutputFile')
            # copy local_file_name to output_file_name
            print(f'Writing output {self.name} to {self.output_file_name}')
            shutil.copy(local_file_name, self.output_file_name)
        else:
            if self.name is None:
                raise Exception('Unexpected: name is None in OutputFile')
            if self.job_id is None:
                raise Exception('Unexpected: job_id is None in OutputFile')
            if self.job_private_key is None:
                raise Exception('Unexpected: job_private_key is None in OutputFile')
            upload_url = _get_upload_url_for_output_file(name=self.name, job_id=self.job_id, job_private_key=self.job_private_key)

            # Upload the file to the URL
            print(f'Uploading output file {self.name}') # it could be a security issue to provide the url in this print statement
            with open(local_file_name, 'rb') as f:
                if not using_mock():
                    resp_upload = requests.put(upload_url, data=f, timeout=60 * 60 * 24 * 7)
                    if resp_upload.status_code != 200:
                        print(upload_url)
                        raise SetOutputFileException(f'Error uploading file to bucket ({resp_upload.status_code}) {resp_upload.reason}: {resp_upload.text}')

        self.was_uploaded = True

    # validator is needed to be an allowed pydantic type
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, cls):
            return value
        else:
            raise ValueError(f'Unexpected type for OutputFile: {type(value)}')



from typing import Union
import os
import requests
import h5py
from pydantic import BaseModel
import remfile


class InputFileDownloadError(Exception):
    pass


class InputFile(BaseModel):
    name: Union[str, None] = None # the name of the input within the context of the processor (not needed when one of url or local_file_name is specified directly)
    url: Union[str, None] = None
    local_file_name: Union[str, None] = None
    job_id: Union[str, None] = None
    job_private_key: Union[str, None] = None

    def get_url(self) -> str:
        if self.url is not None:
            if self.local_file_name is not None:
                raise Exception('Cannot specify both url and local_file_name in InputFile')
            if self.job_id is not None:
                raise Exception('Cannot specify both url and job_id in InputFile')
            if self.job_private_key is not None:
                raise Exception('Cannot specify both url and job_private_key in InputFile')
            return self.url
        elif self.local_file_name is not None:
            if self.job_id is not None:
                raise Exception('Cannot specify both local_file_name and job_id in InputFile')
            if self.job_private_key is not None:
                raise Exception('Cannot specify both local_file_name and job_private_key in InputFile')
            raise Exception('Cannot get url for local_file in InputFile')
        else:
            if self.job_id is None:
                raise Exception('Unexpected: job_id is None')
            if self.job_private_key is None:
                raise Exception('Unexpected: job_private_key is None')
            from .Job import _get_download_url_for_input_file # avoid circular import
            if self.name is None:
                raise Exception('Unexpected: name is None in InputFile')
            return _get_download_url_for_input_file(name=self.name, job_id=self.job_id, job_private_key=self.job_private_key)

    def download(self, dest_file_path: str):
        if self.local_file_name is not None:
            # In the case of a local file, we just copy it
            print(f'Copying {self.local_file_name} to {dest_file_path}')
            import shutil
            shutil.copyfile(self.local_file_name, dest_file_path)
            return
        url = self.get_url()
        print(f'Downloading {url} to {dest_file_path}')
        # stream the download
        r = requests.get(url, stream=True, timeout=60 * 60 * 24 * 7)
        if r.status_code != 200:
            raise InputFileDownloadError(f'Error downloading file {url}: {r.status_code} {r.reason}')
        with open(dest_file_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)

    def get_file(self, *, download: bool = False):
        if self.local_file_name is not None:
            # In the case of a local file, we just return a file object
            f = open(self.local_file_name, 'rb')
            # An issue here is that this file is never closed. Not sure how to fix that.
            return f

        if download:
            if not os.path.exists('_download_inputs'):
                os.mkdir('_download_inputs')
            tmp_fname = f'_download_inputs/{self.name}'
            self.download(tmp_fname)
            f = open(tmp_fname, 'rb')
            # An issue here is that this file is never closed. Not sure how to fix that.
            return f

        # self has a get_url() method
        # It's important to do it this way so that the url can renew as needed
        return remfile.File(self)

    def get_h5py_file(self):
        remf = self.get_file()
        return h5py.File(remf, 'r')

    # validator is needed to be an allowed pydantic type
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, cls):
            return value
        else:
            raise ValueError(f'Unexpected type for InputFile: {type(value)}')



from typing import Union, List
from .... import BaseModel
from fastapi import APIRouter, Header
from ...services._remove_detached_files_and_jobs import _remove_detached_files_and_jobs
from ....common.dendro_types import DendroFile
from ._authenticate_gui_request import _authenticate_gui_request
from ...core._get_project_role import _check_user_can_edit_project
from ...clients.db import fetch_file, fetch_project_files, fetch_project, delete_file as db_delete_file
from ...services.gui.set_file import set_file as service_set_file
from ..common import api_route_wrapper
from ...core._create_random_id import _create_random_id
from ...services.processor.get_upload_url import _get_upload_url_for_object_key

router = APIRouter()

# get file
class GetFileResponse(BaseModel):
    file: DendroFile
    success: bool

@router.get("/projects/{project_id}/files/{file_name:path}")
@api_route_wrapper
async def get_file(project_id, file_name):
    file = await fetch_file(project_id, file_name)
    assert file is not None, f"No file with name {file_name} in project with ID {project_id}"
    return GetFileResponse(file=file, success=True)

# get files
class GetFilesResponse(BaseModel):
    files: List[DendroFile]
    success: bool

@router.get("/projects/{project_id}/files")
@api_route_wrapper
async def get_files(project_id):
    files = await fetch_project_files(project_id)
    return GetFilesResponse(files=files, success=True)

# set file
class SetFileRequest(BaseModel):
    content: str # for example, url:https://...
    jobId: Union[str, None] = None
    size: Union[int, None] = None
    metadata: dict = {}
    isFolder: bool = False

class SetFileResponse(BaseModel):
    fileId: str
    success: bool

@router.put("/projects/{project_id}/files/{file_name:path}")
@api_route_wrapper
async def set_file(project_id, file_name, data: SetFileRequest, github_access_token: str = Header(...)):
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    # parse the request
    content = data.content
    job_id = data.jobId
    size = data.size
    metadata = data.metadata
    is_folder = data.isFolder

    assert size is not None, "size must be specified"

    project = await fetch_project(project_id)
    assert project is not None, f"No project with ID {project_id}"

    _check_user_can_edit_project(project, user_id)

    file_id = await service_set_file(project_id=project_id, user_id=user_id, file_name=file_name, content=content, job_id=job_id, size=size, metadata=metadata, is_folder=is_folder)

    return SetFileResponse(fileId=file_id, success=True)

# delete file
class DeleteFileResponse(BaseModel):
    success: bool

@router.delete("/projects/{project_id}/files/{file_name:path}")
@api_route_wrapper
async def delete_file(project_id, file_name, github_access_token: str = Header(...)):
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    project = await fetch_project(project_id)
    assert project is not None, f"No project with ID {project_id}"

    _check_user_can_edit_project(project, user_id)

    await db_delete_file(project_id, file_name)

    # remove detached files and jobs
    await _remove_detached_files_and_jobs(project_id)

    return DeleteFileResponse(success=True)

# initiate file upload
class CreateFileAndInitiateUploadRequest(BaseModel):
    fileName: str # file name within the project
    size: int
    metadata: dict = {}

class CreateFileAndInitiateUploadResponse(BaseModel):
    fileId: str
    uploadUrl: str
    success: bool

@router.post("/projects/{project_id}/create_file_and_initiate_upload")
@api_route_wrapper
async def create_file_and_initiate_upload(project_id, data: CreateFileAndInitiateUploadRequest, github_access_token: str = Header(...)):
    # Generate a signed URL for uploading a file to the output bucket and set the file in the project.
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    # parse the request
    file_name = data.fileName
    size = data.size
    metadata = data.metadata

    assert size is not None, "size must be specified"

    if size > 100e6:
        raise Exception("File size limit exceeded")

    project = await fetch_project(project_id)
    assert project is not None, f"No project with ID {project_id}"

    _check_user_can_edit_project(project, user_id)

    # it would be nice to use the file name here, but that could cause a problem if the user renames a file, or makes a copy within the project
    _random_file_name = f'{_create_random_id(12)}'
    object_key = f'dendro-uploads/{project_id}/{_random_file_name}'
    upload_url, download_url = await _get_upload_url_for_object_key(object_key, size=size)

    content = f'url:{download_url}'
    file_id = await service_set_file(project_id=project_id, user_id=user_id, file_name=file_name, content=content, job_id=None, size=size, metadata=metadata)

    return CreateFileAndInitiateUploadResponse(fileId=file_id, uploadUrl=upload_url, success=True)



from typing import List
from .InputFile import InputFile
from .InputFolder import InputFolder
from .Job import Job
from .AppProcessor import AppProcessor
from ..common._api_request import _processor_get_api_request


# An empty object that we can set attributes on
# The user is going to think it's a pydantic model, but it's not
# We'll at least give them a dict() and a model_dump() method
class ContextObject:
    def __init__(self) -> None:
        self._dendro_attributes = {}
    def _dendro_set_attribute(self, name, value):
        self._dendro_attributes[name] = value
        setattr(self, name, value)
    def _denro_set_attribute_where_name_may_have_dots(self, name, value):
        if '.' not in name:
            self._dendro_set_attribute(name, value)
            return
        parts = name.split('.')
        obj = self
        for part in parts[:-1]:
            if not hasattr(obj, part):
                obj._dendro_set_attribute(part, ContextObject())
            obj = getattr(obj, part)
            assert isinstance(obj, ContextObject), f'Unexpected type for {part}'
        obj._dendro_set_attribute(parts[-1], value)
    def dict(self):
        ret = {}
        for k, v in self._dendro_attributes.items():
            if isinstance(v, ContextObject):
                ret[k] = v.dict()
            else:
                ret[k] = v
        return ret
    def model_dump(self):
        return self.dict()

def _run_job_child_process(*, job_id: str, job_private_key: str, processors: List[AppProcessor]):
    """
    Used internally to actually run the job by calling the processor function.
    If an app image is being used, this will occur within the container.
    """

    # print statements here will end up in the console output for the job
    print(f'[dendro] Running job: {job_id}')

    # Get a job from the remote dendro API
    job: Job = _get_job(job_id=job_id, job_private_key=job_private_key)

    # Find the registered processor and the associated processor function
    processor_name = job.processor_name
    print(f'[dendro] Processor: {processor_name}')

    processor = next((p for p in processors if p._name == processor_name), None)
    if not processor:
        raise Exception(f'Processor not found: {processor_name}')
    if not processor._processor_class:
        raise Exception(f'Processor does not have a processor_class: {processor_name}')
    processor_class = processor._processor_class

    # Assemble the context for the processor function
    print('[dendro] Assembling context')
    context = ContextObject()
    for input in processor._inputs:
        if not input.list:
            # this input is not a list
            print(f'[dendro] Input: {input.name}')
            input_file = next((i for i in job.inputs if i.name == input.name), None)
            if not input_file:
                raise Exception(f'Input not found: {input.name}')
            context._dendro_set_attribute(input.name, input_file)
        else:
            # this input is a list
            print(f'[dendro] Input (list): {input.name}')
            the_list: List[InputFile] = []
            ii = 0
            while True:
                # find a job input of the form <input_name>[ii]
                input_file = next((i for i in job.inputs if i.name == f'{input.name}[{ii}]'), None)
                if input_file is None:
                    # if not found, we must be at the end of the list
                    break
                the_list.append(input_file)
                ii += 1
            print(f'[dendro] Input (list): {input.name} (found {len(the_list)} files)')
            context._dendro_set_attribute(input.name, the_list)
    for input_folder in processor._input_folders:
        if not input_folder.list:
            # this input folder is not a list
            print(f'[dendro] Input folder: {input_folder.name}')
            input_folder_object = next((i for i in job.input_folders if i.name == input_folder.name), None)
            if not input_folder_object:
                raise Exception(f'Input folder not found: {input_folder.name}')
            context._dendro_set_attribute(input_folder.name, input_folder_object)
        else:
            # this input folder is a list
            print(f'[dendro] Input folder (list): {input_folder.name}')
            the_folder_list: List[InputFolder] = []
            ii = 0
            while True:
                # find a job input folder of the form <input_folder_name>[ii]
                input_folder_object = next((i for i in job.input_folders if i.name == f'{input_folder.name}[{ii}]'), None)
                if input_folder_object is None:
                    # if not found, we must be at the end of the list
                    break
                the_folder_list.append(input_folder_object)
                ii += 1
            print(f'[dendro] Input folder (list): {input_folder.name} (found {len(the_folder_list)} folders)')
            context._dendro_set_attribute(input_folder.name, the_folder_list)
    for output in processor._outputs:
        print(f'[dendro] Output: {output.name}')
        output_file = next((o for o in job.outputs if o.name == output.name), None)
        if not output_file:
            raise Exception(f'Output not found: {output.name}')
        context._dendro_set_attribute(output.name, output_file)
    for output_folder in processor._output_folders:
        print(f'[dendro] Output folder: {output_folder.name}')
        output_folder_object = next((o for o in job.output_folders if o.name == output_folder.name), None)
        if not output_folder_object:
            raise Exception(f'Output folder not found: {output_folder.name}')
        context._dendro_set_attribute(output_folder.name, output_folder_object)
    for parameter in processor._parameters:
        job_parameter = next((p for p in job.parameters if p.name == parameter.name), None)
        if job_parameter is None:
            parameter_value = parameter.default
        else:
            parameter_value = job_parameter.value
        print(f'[dendro] Parameter: {parameter.name} = {parameter_value}')
        context._denro_set_attribute_where_name_may_have_dots(parameter.name, parameter_value)

    print('[dendro] Preparing to run processor')
    _set_custom_kachery_storage_backend(job_id=job_id, job_private_key=job_private_key)

    # Run the processor function
    print('[dendro] Running processor')
    processor_class.run(context)

    # Check that all outputs were set
    print('[dendro] Checking outputs')
    for output in processor._outputs:
        output_file = next((o for o in job.outputs if o.name == output.name), None)
        assert output_file is not None, f'Output not found: {output.name}'
        if not output_file.was_uploaded:
            raise Exception(f'Output was not uploaded: {output.name}')

    # Print a message indicating that the job is complete
    print(f'[dendro] Job complete: {job_id}')

def _get_job(*, job_id: str, job_private_key: str) -> Job:
    """Get a job from the dendro API"""
    job = Job(
        job_id=job_id,
        job_private_key=job_private_key
    )
    return job

def _set_custom_kachery_storage_backend(*, job_id: str, job_private_key: str):
    try:
        import kachery_cloud as kcl
    except ImportError:
        # if we don't have kachery installed, then let's not worry about it
        return

    try:
        custom_storage_backend = CustomKacheryStorageBackend(job_id=job_id, job_private_key=job_private_key)
        kcl.set_custom_storage_backend(custom_storage_backend)
    except Exception as e:
        print('WARNING: Problem setting custom kachery storage backend:', e)
        return

class CustomKacheryStorageBackend:
    def __init__(self, *, job_id: str, job_private_key: str):
        self._job_id = job_id
        self._job_private_key = job_private_key
    def store_file(self, file_path: str, *, label: str):
        sha1 = _compute_sha1_of_file(file_path)

        url_path = f'/api/processor/jobs/{self._job_id}/additional_uploads/sha1/{sha1}/upload_url'
        headers = {
            'job-private-key': self._job_private_key
        }
        res = _processor_get_api_request(
            url_path=url_path,
            headers=headers
        )
        upload_url = res['uploadUrl']
        download_url = res['downloadUrl']

        import requests
        with open(file_path, 'rb') as f:
            resp_upload = requests.put(upload_url, data=f, timeout=60 * 60 * 24 * 7)
        if resp_upload.status_code != 200:
            raise Exception(f'Error uploading file to bucket ({resp_upload.status_code}) {resp_upload.reason}: {resp_upload.text}')
        return download_url


def _compute_sha1_of_file(file_path: str):
    import hashlib
    sha1 = hashlib.sha1()
    with open(file_path, 'rb') as f:
        while True:
            chunk = f.read(2**16)
            if not chunk:
                break
            sha1.update(chunk)
    return sha1.hexdigest()



import uuid


def _create_random_id(length: int) -> str:
    # Generate a random UUID
    full_uuid = uuid.uuid4()

    # Convert to a string and take the first [length] characters
    short_id = str(full_uuid).replace('-', '')[:length]

    return short_id



def _model_dump(model, exclude_none=False):
    # handle both pydantic v1 and v2
    if hasattr(model, 'model_dump'):
        return model.model_dump(exclude_none=exclude_none)
    else:
        return model.dict(exclude_none=exclude_none)



from ....common.dendro_types import DendroProject
from ...clients.db import delete_all_files_in_project, delete_all_jobs_in_project, delete_project as db_delete_project


async def delete_project(project: DendroProject):
    await delete_all_files_in_project(project.projectId)
    await delete_all_jobs_in_project(project.projectId)
    await db_delete_project(project.projectId)
    return None



from typing import List
from fastapi import APIRouter
from ..common import api_route_wrapper
from .... import BaseModel
from ....common.dendro_types import DendroProject
from ...clients.db import fetch_files_with_content_string, fetch_project


router = APIRouter()

# find projects
class FindProjectsRequest(BaseModel):
    fileUrl: str

class CreateProjectResponse(BaseModel):
    projects: List[DendroProject]
    success: bool

@router.post("/find_projects")
@api_route_wrapper
async def find_projects(data: FindProjectsRequest) -> CreateProjectResponse:
    files = await fetch_files_with_content_string(f'url:{data.fileUrl}')
    project_ids: List[str] = []
    for file in files:
        if file.projectId not in project_ids:
            project_ids.append(file.projectId)
    projects: List[DendroProject] = []
    for project_id in project_ids:
        p = await fetch_project(project_id)
        if p is not None:
            projects.append(p)
    return CreateProjectResponse(projects=projects, success=True)



import traceback
from fastapi import HTTPException
from functools import wraps


class AuthException(Exception):
    pass

def api_route_wrapper(route_func):
    @wraps(route_func)
    async def wrapper(*args, **kwargs):
        try:
            return await route_func(*args, **kwargs)
        except AuthException as ae:
            raise HTTPException(status_code=401, detail=str(ae))
        except Exception as e:
            traceback.print_exc()
            raise HTTPException(status_code=500, detail=str(e)) from e
    return wrapper



import os
import time
import traceback
from .common import _get_upload_url, _process_is_alive


def console_output_monitor(parent_pid: str):
    """
    Monitor console output of a job.
    """
    job_id = os.environ.get('JOB_ID', None)
    if job_id is None:
        raise KeyError('JOB_ID is not set')
    job_private_key = os.environ.get('JOB_PRIVATE_KEY', None)
    if job_private_key is None:
        raise KeyError('JOB_PRIVATE_KEY is not set')
    console_out_file = os.environ.get('CONSOLE_OUT_FILE', None)
    if console_out_file is None:
        raise KeyError('CONSOLE_OUT_FILE is not set')

    last_upload_timestamp = 0
    overall_timer = time.time()

    while True:
        if not _process_is_alive(parent_pid):
            print(f'Parent process {parent_pid} is no longer alive. Doing final upload.')
            do_upload(
                console_out_file=console_out_file,
                job_id=job_id,
                job_private_key=job_private_key
            )
            print('Exiting.')
            break

        elapsed_since_upload = time.time() - last_upload_timestamp
        overall_elapsed = time.time() - overall_timer
        if overall_elapsed < 60:
            interval = 10
        elif overall_elapsed < 60 * 5:
            interval = 30
        elif overall_elapsed < 60 * 20:
            interval = 60
        else:
            interval = 120
        if elapsed_since_upload >= interval:
            last_upload_timestamp = time.time()
            do_upload(
                console_out_file=console_out_file,
                job_id=job_id,
                job_private_key=job_private_key
            )

        time.sleep(1)

def do_upload(*, console_out_file, job_id, job_private_key):
    import requests
    if not os.path.exists(console_out_file):
        return True
    with open(console_out_file, 'r') as f:
        console_text = f.read()
    console_text_lines = console_text.split('\n')
    new_lines = []
    for line in console_text_lines:
        # handle carriage return (e.g. in progress bar)
        # find the last \r
        last_cr_index = line.rfind('\r')
        if last_cr_index != -1:
            new_lines.append(line[last_cr_index + 1:].encode('utf-8'))
        else:
            new_lines.append(line.encode('utf-8'))
    text_to_upload = b'\n'.join(new_lines)
    try:
        console_output_upload_url = _get_upload_url(job_id=job_id, job_private_key=job_private_key, output_name='_console_output')
        r = requests.put(console_output_upload_url, data=text_to_upload, timeout=5)
        if r.status_code != 200:
            raise Exception(f'Error uploading console output: {r.status_code} {r.text}')
        return True
    except: # noqa
        print('Error uploading console output')
        traceback.print_exc()
        return False
